import re
import os
import sys
import time
import shutil
import argparse
from io import BytesIO
from pathlib import Path
from collections import deque
from abc import ABC, abstractmethod
from typing import Iterator, List

import requests
from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from rich.layout import Layout
from rich.progress import Progress, BarColumn, TextColumn

from isaac_toolkit.utils.compression import extract_zstd

RV_BASE_CONTAINER_ROOT = "/tools/etiss_arch_riscv/rv_base"
# TODO MGCLIENT CDFGPASS


class RetargetClient(ABC):

    def __init__(self, api_url: str, token: str = None, console: Console = None):
        """
        api_url: base URL, e.g., "http://localhost:8080"
        token: optional API key / token if known
        console: rich.Console instance for live progress
        """
        self.api_url = api_url.rstrip("/")
        self.token = token
        self.console = console

    def _rewrite_cdsl(self, cdsl_path: str) -> bytes:
        """
        Rewrites rv_base imports so they point to the container path.
        """

        cdsl_filename = Path(cdsl_path).name
        text = Path(cdsl_path).read_text()

        import_re = re.compile(r'import\s+"([^"]+)"')

        def replace(match):
            path = match.group(1)

            if "rv_base" in path:
                new_path = path.split("rv_base", 1)[1].lstrip("/")
                return f'import "{RV_BASE_CONTAINER_ROOT}/{new_path}"'

            return match.group(0)

        rewritten = import_re.sub(replace, text)

        return cdsl_filename, rewritten.encode("utf-8")

    def _submit_job(self, files: dict = None, data: dict = None):
        resp = requests.post(f"{self.api_url}/jobs", files=files, data=data)
        resp.raise_for_status()
        data = resp.json()
        self.token = data["token"]
        return data["job_id"], data["token"]

    @abstractmethod
    def submit_job(self, tag: str, *args, **kwargs):
        raise NotImplementedError

    def get_status(self, job_id: str):
        resp = requests.get(f"{self.api_url}/jobs/{job_id}", params={"token": self.token})
        resp.raise_for_status()
        return resp.json()

    def download_artifact(self, job_id: str, dest_path: str):
        resp = requests.get(f"{self.api_url}/jobs/{job_id}/artifact", params={"token": self.token}, stream=True)
        resp.raise_for_status()
        with open(dest_path, "wb") as f:
            for chunk in resp.iter_content(1024 * 1024):
                f.write(chunk)

    def get_log_buf(self, job_id: str, n_lines: int = 50):
        resp = requests.get(
            f"{self.api_url}/jobs/{job_id}/logs", params={"token": self.token, "n_lines": n_lines, "stream": False}
        )
        resp.raise_for_status()
        return resp.text

    def stream_logs(self, job_id: str, n_lines: int = 0) -> Iterator[str]:
        """
        Stream logs line by line (tail -f style).
        """
        with requests.get(
            f"{self.api_url}/jobs/{job_id}/logs",
            params={"token": self.token, "n_lines": n_lines, "stream": True},
            stream=True,
        ) as resp:
            resp.raise_for_status()
            for line in resp.iter_lines(decode_unicode=True):
                if line:
                    yield line

    def wait_for_completion(self, job_id: str, poll_interval: float = 5.0, verbose: bool = False):
        """Waits for job completion and prints live progress"""
        last_progress = -1
        n_lines = 50
        # if USE_RICH:
        # from rich import inspect
        # inspect(console)
        buffer = deque(maxlen=n_lines)
        buffer.append("Waiting for job to start... (queued)")
        # buffer = []
        progress = Progress(
            TextColumn("[bold blue]{task.description}"),
            BarColumn(),
            TextColumn("{task.percentage:>3.0f}%"),
            TextColumn("Elapsed: {task.fields[elapsed_seconds]:.1f}s"),
            TextColumn("ETA: {task.fields[eta_seconds]:.1f}s"),
            # TimeRemainingColumn(),
        )

        task = progress.add_task(f"Running job {job_id}", total=100, elapsed_seconds=0, eta_seconds=0)

        layout = Layout()
        layout.split_column(Layout(progress, size=3), Layout(name="log"))

        def render() -> Layout:
            # console.print("render")
            layout["log"].update(Panel("\n".join(buffer), title=f"Last {n_lines} Lines"))
            return layout

        # TODO: add dummy context if disabled
        #  with Live(render, console=console, refresh_per_second=2):
        assert self.console is not None
        with Live(render(), console=self.console, refresh_per_second=2):
            log_stream = self.stream_logs(job_id, n_lines=n_lines)
            while True:
                status = self.get_status(job_id)
                # if verbose:
                #     assert USE_RICH
                #     if status["status"] in ("finished", "running"):
                #         # log_buf = self.get_log_buf(job_id, n_lines=n_lines)
                #         buffer = log_buf.splitlines()
                #         print("===!!!===")
                #         print("log_buf", log_buf)
                #         print("===???===")
                prog = status.get("progress", 0.0)
                # console.print("prog", prog)
                prog_percent = int(prog * 100)
                # if prog_percent != last_progress:
                if prog_percent >= last_progress:
                    # if USE_RICH:
                    if True:
                        elapsed_seconds = status.get("elapsed_seconds", 0.0)
                        eta_seconds = status.get("eta_seconds", 0.0)
                        progress.update(
                            task, completed=prog_percent, elapsed_seconds=elapsed_seconds, eta_seconds=eta_seconds
                        )
                    else:
                        sys.stdout.write(f"\rJob {job_id} progress: {prog_percent}%")
                        sys.stdout.flush()
                    last_progress = prog_percent
                try:
                    # console.print("try")
                    # for _ in range(5):  # fetch a few lines at a time
                    # for _ in range(n_lines):  # fetch a few lines at a time
                    while True:
                        line = next(log_stream)
                        # console.print("line", line)
                        buffer.append(line)
                        # layout["log"].update(
                        #     Panel("\n".join(buffer), title=f"Last {n_lines} Lines")
                        # )
                except StopIteration:
                    # console.print("stop")
                    pass
                render()
                if status["status"] in ("finished", "failed"):
                    self.console.print(f"\nJob {job_id} completed with status: {status['status']}")
                    error = status.get("error")
                    if error:
                        self.console.print(f"ERROR: {error}")
                    break
                time.sleep(poll_interval)
        return status["status"]

    def download_logs(self, job_id: str, dest_path: str):

        resp = requests.get(f"{self.api_url}/jobs/{job_id}/logs", params={"token": self.token}, stream=True)

        resp.raise_for_status()

        with open(dest_path, "wb") as f:
            for chunk in resp.iter_content(1024 * 1024):
                f.write(chunk)


class Seal5RetargetClient(RetargetClient):

    def submit_job(self, tag: str, cdsl_path: str = None, config_path: str = None, **kwargs):
        assert cdsl_path is not None
        assert config_path is not None
        cdsl_filename, rewritten_cdsl = self._rewrite_cdsl(cdsl_path)
        files = {"cdsl": (dsl_filename, BytesIO(rewritten_cdsl)), "config": open(config_path, "rb")}
        data = {
            "tag": tag,
        }
        return self._submit_job(files=files, data=data)


class EtissRetargetClient(RetargetClient):

    def submit_job(self, tag: str, cdsl_path: str = None, **kwargs):
        assert cdsl_path is not None
        dsl_filename, rewritten_cdsl = self._rewrite_cdsl(dsl_path)
        files = {
            "cdsl": (dsl_filename, BytesIO(rewritten_cdsl)),
        }
        data = {
            "tag": tag,
        }
        return self._submit_job(files=files, data=data)


class EtissPerfRetargetClient(RetargetClient):

    def submit_job(self, tag: str, cdsl_path: str = None, cpdsl_path: str = None, **kwargs):
        assert cdsl_path is not None
        assert cpdsl_path is not None
        cdsl_filename, rewritten_cdsl = self._rewrite_cdsl(cdsl_path)
        cpdsl_filename = Path(cpdsl_path).name
        files = {
            "cdsl": (cdsl_filename, BytesIO(rewritten_cdsl)),
            "cpdsl": (cpdsl_filename, cpdsl_path),
        }
        data = {
            "tag": tag,
        }
        return self._submit_job(files=files, data=data)


def run_retargeting_service(
    files: List[str],
    out_dir: str = None,
    verbose: bool = False,
    host: str = "localhost",
    port: int = 8080,
    tag: str = None,
):

    assert out_dir is not None
    assert tag is not None
    out_path = Path(out_dir)
    out_path.mkdir(exist_ok=True)
    console = Console(force_terminal=True, force_interactive=True) if verbose else None

    client_lookup = {
        "llvm": Seal5RetargetClient,
        "iss": EtissRetargetClient,
        "perf": EtissPerfRetargetClient,
    }
    client_cls = client_lookup.get(tag)
    assert client_cls is not None, f"Unable to find client class for tag: {tag}"

    client = RetargetClient(f"http://{host}:{port}", console=console)

    # Separate input files
    cdsl_files: List[Path] = []
    cpdsl_files: List[Path] = []
    cfg_files: List[Path] = []
    for f in files:
        file_path = Path(f)
        if not file_path.is_file():
            raise ValueError(f"Not a file: {f}")
        if file_path.suffix in [".core_desc", ".cdsl"]:
            cdsl_files.append(file_path)
        elif file_path.suffix in [".core_perf_dsl"]:
            cpdsl_files.append(file_path)
        elif file_path.suffix in [".yaml", ".yml"]:
            cfg_files.append(file_path)
    if len(cdsl_files) != 1:
        raise ValueError(f"Expected exactly one .core_desc/.cdsl file, got {len(cdsl_files)}")
    cdsl_file = cdsl_files[0]
    cpdsl_file = None
    if len(cpdsl_files) > 0:
        if len(cpdsl_files) > 1:
            raise ValueError(f"Expected exactly one .core_perf_dsl file, got {len(cpdsl_files)}")
        cpdsl_file = cpdsl_files[0]

    # Merge configs if multiple
    if len(cfg_files) > 1:
        merged_contents = "\n".join([f.read_text() for f in cfg_files]).replace("---\n", "\n")
        merged_cfg_file = out_path / "merged_config.yml"
        merged_cfg_file.write_text(merged_contents)
        cfg_file = merged_cfg_file
    else:
        cfg_file = cfg_files[0]

    job_id, token = client.submit_job(tag, cdsl_path=str(cdsl_file), cfg_path=str(cfg_file), cpdsl_path=str(cpdsl_file))

    if console:
        console.print(f"Submitted job {job_id} with token {token}")
    time.sleep(1)

    # Wait for completion
    status = client.wait_for_completion(job_id) if console else None

    # Download logs
    log_file = out_path / f"{job_id}_logs.txt"
    if console:
        console.print(f"Downloading logs to {log_file}")
    client.download_logs(job_id, str(log_file))

    if status == "failed":
        if console:
            console.print("Artifact not available (job failed)")
    else:
        # Download artifact
        artifact_file = out_path / f"{job_id}_artifact.tar.zst"
        if console:
            console.print(f"Downloading artifact to {artifact_file}")
        client.download_artifact(job_id, str(artifact_file))
        if console:
            console.print("Download finished")
            console.print(f"Extracting artifact to {out_path}")
        extract_zstd(artifact_file, out_path)

        # Move files
        shutil.copytree(out_path / "output", out_path, dirs_exist_ok=True)
        shutil.rmtree(out_path / "output")
        if console:
            console.print("Done")
            console.print(f"Removing archive {artifact_file}")
        os.remove(artifact_file)
        if console:
            console.print("Done")


def handle(args):
    # assert args.session is not None
    sess = None
    if args.session is not None:
        session_dir = Path(args.session)
        assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
        sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    run_retargeting_service(
        args.files,
        out_dir=args.out_dir,
        force=args.force,
        verbose=args.verbose,
        host=args.host,
        port=args.port,
        tag=args.tag,
    )
    if sess is not None:
        sess.save()


def get_parser():
    parser = argparse.ArgumentParser(description="Submit jobs to Retarget server and fetch results")
    parser.add_argument("files", nargs="+", help="Input files (.core_desc/.cdsl and config YAML)")
    parser.add_argument("--tag", required=True, help="Job tag")
    parser.add_argument("--output", "-o", dest="out_dir", help="Output directory")
    parser.add_argument("--session", "--sess", "-s", type=str, required=False)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--host", default=os.environ.get("HOST", "localhost"), help="Server host override")
    parser.add_argument("--port", default=os.environ.get("PORT", "8080"), help="Server port override")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--verbose", action="store_true", help="Enable verbose console output")
    args = parser.parse_args()

    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



import sys
import pandas as pd

assert len(sys.argv) in [2, 3]

csv_file = sys.argv[1]
output_file = None if len(sys.argv) == 2 else sys.argv[2]

# Load CSV
df = pd.read_csv(csv_file)

# Pivot to get start and end times
# Assuming consecutive rows are start and end of each stage
gantt = []
gantt.append("  section Run 0")
stages = df["stage"].tolist()
times = df["time_ms"].tolist()
current = 0
for i in range(0, len(stages)):
    stage_name = stages[i]
    start = current
    end = start + times[i]
    current = end
    gantt.append(f"    {stage_name} Stage : {start}, {end}")

content = ""
content += "```mermaid\n"
content += "gantt\n"
content += "  title Flow\n"
content += "  dateFormat x\n"
content += "  axisFormat %H:%M:%S\n"
content += "\n".join(gantt)
content += "\n```"

# Write Mermaid Gantt
if output_file is None:
    print(content)
else:
    with open(output_file, "w") as f:
        f.write(content)
    print(f"Mermaid Gantt chart written to {output_file}")



#
# Copyright (c) 2024 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# # import os
import sys
import shutil
import subprocess

# import yaml
import logging
import argparse
from typing import Optional, Union
from pathlib import Path

from isaac_toolkit.session import Session


DEFAULT_CORE = "CVA5"
SUPPORTED_CORES = ["CVA5", "VEX"]

DEFAULT_PDK = "nangate45"
SUPPORTED_PDKS = ["nangate45"]

DEFAULT_CLK_PERIOD = 50.0
PER_CORE_CLK_PERIOD = {"CVA5": 50.0}


logger = logging.getLogger("synopsys")


def run_synopsys_asip_syn(
    sess: Session,
    workdir: Optional[Union[str, Path]] = None,
    label: Optional[str] = None,
    core_name: Optional[str] = None,
    pdk: Optional[str] = None,
    clk_period: Optional[float] = None,
    script_path: Optional[Union[str, Path]] = None,
    constraints_file: Optional[Union[str, Path]] = None,
    dse: bool = False,
    force: bool = False,
):
    if dse:
        raise NotImplementedError
    assert workdir is not None
    if not isinstance(workdir, Path):
        workdir = Path(workdir)
    assert workdir.is_dir()
    if core_name is None:
        core_name = DEFAULT_CORE
    assert core_name in SUPPORTED_CORES
    assert label is not None
    assert script_path is not None
    if pdk is None:
        pdk = DEFAULT_PDK
    assert pdk in SUPPORTED_PDKS
    if clk_period is None:
        clk_period = PER_CORE_CLK_PERIOD.get(core_name, DEFAULT_CLK_PERIOD)
    assert constraints_file is not None
    assert Path(constraints_file).is_file()

    possible_subdirs = ["docker", "local"]
    for possible_subdir in possible_subdirs:
        base_dir = workdir / possible_subdir
        hls_dir = base_dir / "hls" / label
        if hls_dir.is_dir():
            break
    assert hls_dir.is_dir()
    output_dir = base_dir / "asip_syn" / label
    if output_dir.is_dir():
        assert force, f"Directory already exists: {output_dir}. Use --force or different --label."
        logger.info("Cleaning up old output dir: %s (--force)", output_dir)
        shutil.rmtree(output_dir)
    else:
        output_dir.mkdir(parents=True)
    hls_rtl_dir = hls_dir / "rtl"
    rtl_dir = output_dir / "rtl"
    shutil.copytree(hls_rtl_dir, rtl_dir)
    script_args = [
        script_path,
        output_dir,
        rtl_dir,
        core_name,
        pdk,
        clk_period,
        constraints_file,
    ]
    subprocess.run(script_args, check=True)


def handle(args):
    # assert args.session is not None
    sess = None
    if args.session is not None:
        session_dir = Path(args.session)
        assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
        sess = Session.from_dir(session_dir)
    run_synopsys_asip_syn(
        sess,
        force=args.force,
        workdir=args.workdir,
        tools_dir=args.tools_dir,
        docker_image=args.docker,
        isaxes=args.isaxes.split(";") if args.isaxes is not None else None,
        index_files=args.index.split(";") if args.index is not None else None,
    )
    if sess is not None:
        sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    # parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--session", "--sess", "-s", type=str, required=False)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--workdir", type=str, default=None)
    parser.add_argument("--dse", action="store_true")
    # label: Optional[str] = None,
    # core_name: Optional[str] = None,
    # pdk: Optional[str] = None,
    # clk_period: Optional[float] = None,
    # script_path: Optional[str, Path] = None,
    # constraints_file: Optional[str, Path] = None,

    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import io
import sys
import leb128
import argparse
from pathlib import Path
from collections import defaultdict

import pandas as pd
from elftools.elf.elffile import ELFFile
from elftools.elf.sections import SymbolTableSection

from capstone import Cs, CS_ARCH_RISCV, CS_MODE_RISCV32, CS_MODE_RISCV64, CS_MODE_RISCVC

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


def parse_elf(elf_path):
    GISEL = True  # TODO: make arg
    func_to_addrs = defaultdict(list)
    with open(elf_path, "rb") as f:
        elffile = ELFFile(f)

        # extract disassembly to count instructions per bb
        # see https://isleem.medium.com/create-your-own-disassembler-in-python-pefile-capstone-754f863b2e1c
        code = elffile.get_section_by_name(".text")
        ops = code.data()
        begin = code["sh_addr"]
        sz = code["sh_size"]
        end = begin + sz
        xlen = elffile.elfclass
        assert xlen is not None

        mode = CS_MODE_RISCV32 if xlen == 32 else CS_MODE_RISCV64
        md = Cs(CS_ARCH_RISCV, mode | CS_MODE_RISCVC)
        # md = Cs(CS_ARCH_RISCV, CS_MODE_RISCV32)
        # valid_pcs = set(x.address for x in md.disasm(ops, addr))
        valid_pcs = set()
        while True:
            temp = [(x.address, x.size) for x in md.disasm(ops, begin)]
            valid_pcs_ = set(pc for pc, size in temp)
            last = temp[-1]
            last_pc, last_size = last
            begin = max(last_pc, begin) + last_size
            valid_pcs.update(valid_pcs_)
            if begin >= end:
                break

        section = elffile.get_section_by_name(".symtab")
        addr_bytes = int(xlen / 8)

        if not section:
            pass
            # print('  No symbol table found. Perhaps this ELF has been stripped?')  # TODO: make err?

        if isinstance(section, SymbolTableSection):
            num_symbols = section.num_symbols()
            # print("  It's a symbol section with %s symbols" % num_symbols)
            for i in range(num_symbols):
                sym = section.get_symbol(i)
                if sym["st_info"]["type"] != "STT_FUNC":
                    continue
                # print(i, sym.entry, sym.name)
                name = sym.name
                # print("name", name)
                # assert name not in func_to_addr, f"Conflict! {name} already in map: {func_to_addr}"
                func_to_addrs[name].append((sym["st_value"], sym["st_size"]))

        addr_to_func = {}
        aliases = defaultdict(list)
        for func_name, funcs in func_to_addrs.items():
            for func in funcs:
                func_addr, func_sz = func
                # print(func_name, ":", func_sz,  hex(func_addr), "-", hex(func_addr + func_sz))
                if func_addr in addr_to_func:
                    if addr_to_func[func_addr] != func_name:
                        aliases[func_name].append(addr_to_func[func_addr])
                else:
                    addr_to_func[func_addr] = func_name
        llvm_bb_addr_map_raw = elffile.get_section_by_name(".llvm_bb_addr_map")
        assert llvm_bb_addr_map_raw is not None, "Missing: .llvm_bb_addr_map"
        llvm_bb_addr_map_raw = llvm_bb_addr_map_raw.data()

        def decode_map(data, addr_to_func):
            ret = {}
            unknown_count = 0
            with io.BytesIO(data) as reader:
                while True:
                    # print("reader", reader)
                    # print("dir(reader)", dir(reader))
                    version = int.from_bytes(reader.read(1), byteorder="little")
                    # print("version", version)
                    # assert version == 2
                    if version != 2:
                        if version == 0:
                            break
                        raise NotImplementedError(f"LLVM_BB_ADDR_MAP version {version}")
                        # print("!")
                        # print(reader.read(100))
                    features = int.from_bytes(reader.read(1), byteorder="little")
                    # print("features", features)
                    assert features == 0
                    func_addr = int.from_bytes(reader.read(addr_bytes), byteorder="little")
                    # print("func_addr", func_addr)
                    func_name = addr_to_func.get(func_addr, None)
                    # print("func_name", func_name)
                    # assert func_name is not None
                    if func_name is None:
                        func_name = f"unknown_func_{unknown_count}"
                        unknown_count += 1
                    # TODO: leb128?
                    # num_bbs = int.from_bytes(reader.read(1), byteorder="little")
                    num_bbs = leb128.u.decode_reader(reader)[0]
                    # print("num_bbs", num_bbs)
                    # assert num_bbs > 0
                    if GISEL:
                        tmp = {}
                    else:
                        tmp = [None] * num_bbs
                    cur = func_addr
                    for i in range(num_bbs):
                        # print("i", i)
                        # bb_id = int.from_bytes(reader.read(1), byteorder="little")
                        bb_id = leb128.u.decode_reader(reader)[0]
                        if not GISEL:
                            assert bb_id < num_bbs
                        start_offset = leb128.u.decode_reader(reader)[0]
                        # print("start_offset", start_offset)
                        assert start_offset >= 0
                        end_offset = leb128.u.decode_reader(reader)[0]
                        # print("end_offset", end_offset)
                        assert end_offset >= 0
                        # TODO: leb128?
                        # metadata = int.from_bytes(reader.read(1), byteorder="little")
                        # metadata = leb128.u.decode_reader(reader)[0]
                        _ = leb128.u.decode_reader(reader)[0]
                        # print("metadata", metadata)
                        # TODO: decode metadata (is_return, is_call, ...)
                        # rest = reader.read(1000)
                        # print("rest", rest)
                        cur += start_offset
                        sz = end_offset - start_offset
                        assert sz >= 0
                        start = cur
                        end = cur + sz
                        if sz > 0:
                            pcs = [pc for pc in range(start, end, 2) if pc in valid_pcs]
                            # valid_pcs_sorted = sorted(list(valid_pcs))
                            # end_pc_idx = valid_pcs_sorted.index(pcs[-1])
                            # end_pc = valid_pcs_sorted[end_pc_idx]
                            # end_pc_next = valid_pcs_sorted[end_pc_idx+1]
                            # end_pc_next_next = valid_pcs_sorted[end_pc_idx+2]
                            num_instrs = len(pcs)
                        else:
                            num_instrs = 0
                        cur += sz
                        if GISEL:
                            tmp[str(bb_id)] = (start, end, sz, num_instrs)
                        else:
                            tmp[bb_id] = (start, end, sz, num_instrs)
                    ret[func_name] = tmp
            return ret

        llvm_bb_addr_map = decode_map(llvm_bb_addr_map_raw, addr_to_func)
        # print("llvm_bb_addr_map", llvm_bb_addr_map)
        VERBOSE = False
        if VERBOSE:
            for func_name in func_to_addrs.keys():
                print(f"{func_name}:")
                bbs = llvm_bb_addr_map.get(func_name, None)
                # PRINT_MISSING = False
                PRINT_MISSING = True
                if bbs is None:
                    if PRINT_MISSING:
                        logger.debug("> no bb addr info found")
                    continue
                if GISEL:
                    bbs = dict(sorted(bbs.items(), key=lambda x: int(x[0]))).values()
                for i, bb in enumerate(bbs):
                    start, end, sz, num_instrs = bb
                    logger.debug(
                        f"> bb{i}",
                        ":",
                        hex(start),
                        "-",
                        hex(end),
                        f"(len={sz}B, num={num_instrs})",
                    )
                    # print()
    return llvm_bb_addr_map


def analyze_llvm_bbs(sess: Session, force: bool = False):
    logger.info("Analyzing LLVM BBs...")
    artifacts = sess.artifacts
    # print("artifacts", artifacts)
    elf_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.ELF)
    # print("elf_artifacts", elf_artifacts)
    assert len(elf_artifacts) == 1
    elf_artifact = elf_artifacts[0]
    # trace_pc2bb_artifacts = filter_artifacts(
    #     artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "pc2bb"
    # )  # TODO: optional or different pass
    # if len(trace_pc2bb_artifacts) > 0:
    #     assert len(trace_pc2bb_artifacts) == 1
    #     trace_pc2bb_artifact = trace_pc2bb_artifacts[0]
    #     # trace_pc2bb_df = trace_pc2bb_artifact.df
    # else:
    #     pass
    #     # trace_pc2bb_df = None

    llvm_bbs = parse_elf(elf_artifact.path)
    # print("llvm_bbs", llvm_bbs)
    df_data = []
    for func_name, func_data in llvm_bbs.items():
        # print("fn", func_name)
        for bb_name, bb_data in func_data.items():
            # print("bn", bb_name)
            bb_name = f"%bb.{bb_name}"
            start, end, sz, num_instrs = bb_data
            new = {
                "func_name": func_name,
                "bb_name": bb_name,
                "pcs": (start, end),
                "size": sz,
                "num_instrs": num_instrs,
            }
            # if trace_pc2bb_df is not None:
            #     print("trace_pc2bb_df", trace_pc2bb_df)
            #     # trace_pc2bb_df[["start", "end"]] = trace_pc2bb_df["bb"].apply(pd.Series)
            #     # matches = trace_pc2bb_df.where(lambda x: x["start"] >= start and x["end"] <= end)
            #     matches = trace_pc2bb_df.where(lambda x: x["start"] >= start).dropna()
            #     matches = matches.where(lambda x: x["end"] <= end).dropna()
            #     if len(matches) > 0:
            #         print("matches", matches)
            #         weights = matches["weight"].sum()
            #         rel_weights = matches["rel_weight"].sum()
            #         # print("weights", weights)
            #         # print("rel_weights", rel_weights)
            #         # input("p")
            #         new["num_trace_bbs"] = len(matches)
            #         new["weight"] = weights
            #         new["rel_weight"] = rel_weights
            #     else:
            #         print("not found")
            #         print("start", start)
            #         print("end", end)
            #         print("sz", sz)
            #         # if start == 85388:
            #         if True:
            #             matches2 = trace_pc2bb_df.where(lambda x: x["start"] == start).dropna()
            #             print("matches2", matches2)
            #             matches2 = matches2.where(lambda x: x["end"] > end).dropna()
            #             print("matches2_", matches2)
            #             if len(matches2) > 0:
            #                 assert len(matches2) == 1
            #                 idx = matches2.index[0]
            #                 row = matches2.iloc[0]
            #                 start_ = row["start"]
            #                 end_ = row["end"]
            #                 size_ = row["size"]
            #                 freq_ = row["freq"]
            #                 weight_ = row["weight"]
            #                 rel_weight_ = row["rel_weight"]
            #                 num_instrs_ = row["num_instrs"]
            #                 print("idx", idx)
            #                 print("row", row)
            #                 print("start_", start_)
            #                 print("end_", end_)
            #                 print("size_", size_)
            #                 print("freq_", freq_)
            #                 print("weight_", weight_)
            #                 print("rel_weight_", rel_weight_)
            #                 print("num_instrs", num_instrs_)
            #                 default_enc_size = 4  # TODO: do not hardcode
            #                 trace_pc2bb_df.loc[idx, "end"] = end
            #                 trace_pc2bb_df.loc[idx, "size"] = sz
            #                 trace_pc2bb_df.loc[idx, "weight"] = weight_ * (sz / size_)
            #                 trace_pc2bb_df.loc[idx, "rel_weight"] = rel_weight_ * (sz / size_)
            #                 trace_pc2bb_df.loc[idx, "num_instrs"] = sz / default_enc_size
            #                 # new2 = {"start": end + default_enc_size, "end": end_, "freq": freq_,
            #                     "size": size_ - sz, "weight": weight_ * (1 - sz / size_),
            #                     "rel_weight": rel_weight_ * (1 - sz / size_),
            #                     "num_instrs": (size_ - sz) / default_enc_size}
            #                 new2 = {"start": end, "end": end_, "freq": freq_, "size": size_ - sz,
            #                     "weight": weight_ * (1 - sz / size_),
            #                     "rel_weight": rel_weight_ * (1 - sz / size_),
            #                     "num_instrs": (size_ - sz) / default_enc_size}
            #                 trace_pc2bb_df = pd.concat([trace_pc2bb_df, pd.DataFrame([new2])])
            #                 # TODO: export as updated artifact?
            #                 new["num_trace_bbs"] = 1
            #                 new["weight"] = weight_ * (sz / size_)
            #                 new["rel_weight"] = rel_weight_ * (sz / size_)
            #                 print("new", new)
            #                 print("new2", new2)

            #                 # input("ooo")
            #             else:
            #                 matches3 = trace_pc2bb_df.where(lambda x: x["start"] <= start).dropna()
            #                 print("matches3", matches3)
            #                 matches3 = matches2.where(lambda x: x["end"] >= end).dropna()
            #                 print("matches3_", matches3)
            #                 if func_name == "tvmgen_default_fused_nn_contrib_conv2d_NCHWc" and bb_name == "%bb.33":
            #                     input("lll")
            #                 if len(matches3) > 0:
            #                     input("uuu")
            #             # if llvm bbs is shorter than trace bbs, we may have untaken backwards branches in the trace
            #             # to split the trace bf into multiple ones, do a reverse search
            # if (start, end) in trace_pc2bb_df["bb"]:
            #     input("yes")
            df_data.append(new)
    # pc2locs_df = pd.DataFrame(pc2locs.items(), columns=["pc", "locs"])
    llvm_bbs_df = pd.DataFrame(df_data)
    # llvm_bbs_df.sort_values("rel_weight", inplace=True, ascending=False)

    attrs = {
        "elf_file": elf_artifact.name,
        "kind": "mapping",
        "by": "isaac_toolkit.analysis.static.llvm_bbs",
    }

    llvm_bbs_artifact = TableArtifact("llvm_bbs", llvm_bbs_df, attrs=attrs)
    # print("llvm_bbs_artifact", llvm_bbs_artifact)
    sess.add_artifact(llvm_bbs_artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    analyze_llvm_bbs(sess, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import sys
import argparse
from pathlib import Path

from neo4j import GraphDatabase, Query
import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, filter_artifacts
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


def anonotate_helper(
    driver,
    func_name: str,
    bb_name: str,
    num_instrs: int,
    freq: int,
    rel_weight: float,
    label: str = "default",
    check: bool = True,
    timeout: float = 120,
):
    bb_query = Query(
        f"""
    MATCH (a:BB)
    WHERE a.func_name = "{func_name}"
    AND a.name = "{bb_name}"
    AND a.session = "{label}"
    SET a.bb_freq = {freq}
    SET a.bb_instrs = {num_instrs}
    SET a.bb_rel_weight = {rel_weight}
    // RETURN *
    """,
        timeout=timeout,
    )

    session = driver.session()
    try:
        # results = session.run(bb_query)
        _ = session.run(bb_query)
        # print("results", results)
        # print("results.df", results.df)
        # input("<>")
    finally:
        session.close()

    instrs_query = Query(
        f"""
    MATCH (a:INSTR)
    WHERE a.func_name = "{func_name}"
    AND a.basic_block = "{bb_name}"
    AND a.session = "{label}"
    AND a.op_type != "input"
    AND a.op_type != "constant"
    SET a.instr_freq = {freq}
    SET a.instr_rel_weight = {rel_weight / num_instrs}
    RETURN COUNT(a)
    """,
        timeout=timeout,
    )

    session = driver.session()
    try:
        # results = session.run(instrs_query)
        _ = session.run(instrs_query)
        # print("results", results, dir(results))
        if check:
            pass
            # df = results.to_df()
            # count = df.iloc[0, 0]
            # print("count", count)
            # print("num_instrs", num_instrs)
            # assert count <= (num_instrs * 1.05 + 5)
        # input("<>")
    finally:
        session.close()


def annotate_bb_weights(sess: Session, label: str = "default", force: bool = False):
    logger.info("Annotating BB weights...")
    memgraph_config = sess.config.memgraph
    memgraph_host = os.environ.get("MEMGRAPH_HOST")
    memgraph_port = os.environ.get("MEMGRAPH_PORT")
    user = ""
    password = ""
    if memgraph_config is not None:
        memgraph_host = memgraph_host or memgraph_config.hostname
        memgraph_port = memgraph_port or memgraph_config.port
        user = memgraph_config.user
        password = memgraph_config.password

    driver = GraphDatabase.driver(f"bolt://{memgraph_host}:{memgraph_port}", auth=(user, password))
    try:

        artifacts = sess.artifacts
        llvm_bbs_artifacts = filter_artifacts(
            artifacts,
            lambda x: x.flags & ArtifactFlag.TABLE and x.name == "llvm_bbs_new",
        )  # TODO: optional or different pass
        assert len(llvm_bbs_artifacts) == 1
        llvm_bbs_artifact = llvm_bbs_artifacts[0]
        llvm_bbs_df = llvm_bbs_artifact.df
        # print("llvm_bbs_df", llvm_bbs_df, llvm_bbs_df.columns)
        # input("?")
        for index, row in llvm_bbs_df.iterrows():
            # print("index", index)
            # print("row", row)
            func_name = row["func_name"]
            bb_name = row["bb_name"]
            num_instrs = row["num_instrs"]
            rel_weight = row["rel_weight"]
            freq = row["freq"]
            if not num_instrs or pd.isna(num_instrs):
                continue
            if not rel_weight or pd.isna(rel_weight):
                continue
            anonotate_helper(
                driver,
                func_name,
                bb_name,
                num_instrs,
                freq,
                rel_weight,
                label=label,
                check=True,
            )
    finally:
        driver.close()


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    annotate_bb_weights(sess, label=args.label, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--label", default="default", required=True)
    parser.add_argument("--force", "-f", action="store_true")
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Dict, List, Set, Tuple, Optional, Union

import bisect
import time

import sys
import logging
import argparse
import posixpath
from pathlib import Path
from collections import defaultdict

import pandas as pd
import numpy as np

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, filter_artifacts
from isaac_toolkit.arch.riscv import riscv_branch_instrs, riscv_return_instrs
from isaac_toolkit.utils.demangle import unmangle_helper


logging.basicConfig(level=logging.DEBUG)  # TODO
logger = logging.getLogger(__name__)


PC_FUNC_NAME_CACHE = {}


# TODO: reset?


def find_func_name(mapping: Dict[str, Tuple[int, int]], pc: int) -> str:
    # TODO: refactor to use existing tables
    """
    Given a program counter, find the function it belongs to
    """
    found = PC_FUNC_NAME_CACHE.get(pc)
    if found is not None:
        return found
    # print("mapping", mapping)
    for func, ranges in mapping.items():
        for range_ in ranges:
            if pc >= range_[0] and pc <= range_[1]:
                PC_FUNC_NAME_CACHE[pc] = func
                return func
    ret = hex(pc)
    # print("hex")
    # input("???")
    PC_FUNC_NAME_CACHE[pc] = ret
    return ret


# TODO: consistent arg naming
# TODO: re-format quotes


def build_call_trace(function_trace_df, event_names):
    call_records = []
    call_stack = []
    funcs = function_trace_df["func"].to_numpy()
    depths = function_trace_df["depth"].to_numpy()
    costs = function_trace_df["cost"].to_numpy()
    start_pcs = function_trace_df["start_pc"].to_numpy()

    # for _, row in function_trace_df.iterrows():
    for i in range(len(function_trace_df)):
        # func, depth, metrics = row["func"], row["depth"], row["cost"]
        func, depth, cost = funcs[i], depths[i], costs[i]

        if not call_stack or depth > call_stack[-1][1]:
            # print("if")
            # function entry
            if call_stack:
                # print("if2")
                caller, _, call_pc = call_stack[-1]
                call_records.append(
                    {
                        "caller": caller,
                        "callee": func,
                        "call_pc": call_pc,  # optional: store PC if tracked
                        "calls": 1,
                        "cost": cost,
                    }
                )
            call_stack.append((func, depth, start_pcs[i]))
            # print("end")
        else:
            # print("else")
            # function exit or same-level
            while call_stack and call_stack[-1][1] >= depth:
                # print("while")
                call_stack.pop()
            call_stack.append((func, depth, start_pcs[i]))
            # print("end2")

    call_trace_df = pd.DataFrame(call_records)
    print("ctdf", call_trace_df)

    # Aggregate duplicates (same caller/callee/pc)
    agg_dict = {"calls": "sum"}
    # for ev in event_names:
    #     agg_dict[ev] = "sum"

    call_trace_df = call_trace_df.groupby(["caller", "callee", "call_pc"], as_index=False).agg(agg_dict)
    print("ctdf", call_trace_df)

    return call_trace_df


# def callgrind_format_get_inclusive_cost(bbs: List[BasicBlock]):
def callgrind_format_get_inclusive_cost(
    bb_trace_df: pd.DataFrame, unique_bbs_df: pd.DataFrame, event_names: List[str] = [], event_arrays=None
):
    call_stack = []
    bb_stack = []
    # total_cost = 0
    inclusive_cost_dict = defaultdict(lambda: defaultdict(list))
    prev_bb_idx = None
    print("P", time.time())
    print("bb_trace_df.head()", bb_trace_df.head())
    print("unique_bbs_df.head()", unique_bbs_df.head())
    unique_bbs_records = unique_bbs_df.to_records(index=False)
    print("len(unique_bbs_records)", len(unique_bbs_records))
    print("len(unique_bbs_df)", len(unique_bbs_df))

    # call_stack: [A, B]
    # [Given] bb_stack: [[bb1, bb2, bb3], [bb4, bb5, bb6]]
    # [B returns]: [[bb1, bb2, [bb3, # cost from B]]]
    # where bb1 - bb3 belong to func A and bb4 - bb6 belong to func B
    # for i, bb in enumerate(bbs):
    # for i, bb_trace_row in bb_trace_df.iterrows():
    bb_funcs = unique_bbs_df["func"].values  # dtype=object or better: categorical/int
    bb_num_instrs = unique_bbs_df["num_instrs"].values  # dtype=int32/int64
    bb_end_instrs = unique_bbs_df["end_instr"].values  # dtype=object or categorical
    bb_first_pc = unique_bbs_df["first_pc"].values
    bb_last_pc = unique_bbs_df["last_pc"].values
    function_trace = []
    # t0 = time.time()
    # ts = [t0]
    # for i, bb_idx in enumerate(bb_trace_df["bb_idx"].values):
    for i, bb_idx in enumerate(bb_trace_df["bb_idx"].values):
        # print("progress", i/len(bb_trace_df)*100, "%")
        # ts.append(time.time())
        # if i == 100:
        #     break
        func = bb_funcs[bb_idx]
        # n_instr = bb_num_instrs[bb_idx]
        # end_instr = bb_end_instrs[bb_idx]
        first_pc = bb_first_pc[bb_idx]
        last_pc = bb_last_pc[bb_idx]
        # print("progress", float(i)/len(bb_trace_df))
        # bb_idx = bb_trace_row.bb_idx
        # bb = unique_bbs_df.iloc[bb_idx]
        # bb = unique_bbs_records[bb_idx]
        # print("bb_ifx", bb_idx)
        # print("func", func)
        # row = bb_cost_trace_df.iloc[i]
        # print("row", row)
        # metrics = [int(row[ev]) for ev in event_names]
        # print("metrics", metrics)
        # print("i,bb_idx,bb", i, bb_idx, bb)
        # total_cost += n_instr
        # print("bb", bb)
        if prev_bb_idx is None or (bb_end_instrs[prev_bb_idx] in riscv_branch_instrs and bb_funcs[prev_bb_idx] != func):
            # first bb in the trace
            # print("append call_stack + bb_stack")
            call_stack.append(func)
            depth = len(call_stack)
            new_ = [func, depth, first_pc, last_pc, np.zeros(len(event_names), dtype=np.int64)]
            function_trace.append(new_)
            bb_stack.append([bb_idx])
        elif bb_funcs[prev_bb_idx] == func:
            # jalr doesn't necessarily mean return
            # 0x2ac8, jalr, memset -> 0x2ae8, memset
            # print("append bb_stack")
            bb_stack[-1].append(bb_idx)
        # elif prev_bb.end_instr in return_instrs:
        # elif prev_bb.end_instr in riscv_return_instrs:
        elif bb_end_instrs[prev_bb_idx] in riscv_return_instrs or bb_num_instrs[prev_bb_idx] == 1:
            # Check whether jalr refer to return
            # sometimes jalr simply means indirect jump
            # TODO: Redundant? Is it already handled in the above condition?
            if func not in call_stack:
                call_stack.append(func)
                depth = len(call_stack)
                # metrics_vec = [event_arrays[ev][bb_idx] for ev in event_names]
                metrics_vec = event_arrays[:, bb_idx]
                new_ = [func, depth, first_pc, last_pc, metrics_vec]
                function_trace.append(new_)
                bb_stack.append([bb_idx])
                # print("append call_stack + bb_stack 2")
                continue

            # ret case for the rest code snippet
            diff = len(call_stack) - call_stack.index(func) - 1

            for j in range(diff):
                # cost_vec = [0] * len(event_names)
                cost_vec = np.zeros(len(event_names), dtype=np.int64)
                callee_first_bb_idx = bb_stack[-1][0][0] if isinstance(bb_stack[-1][0], list) else bb_stack[-1][0]
                for bb_stack_elem in bb_stack[-1]:
                    # print("bb_stack_elem", bb_stack_elem)
                    if not isinstance(bb_stack_elem, list):
                        # base case: plain BB
                        # bb_metrics = bb_cost_trace_df.iloc[bb_stack_elem]
                        # metrics_vec = [int(bb_metrics[ev]) for ev in event_names]
                        # metrics_vec = [event_arrays[ev][bb_stack_elem] for ev in event_names]
                        metrics_vec = event_arrays[:, bb_stack_elem]
                        # cost_vec = [c + m for c, m in zip(cost_vec, metrics_vec)]
                        cost_vec = cost_vec + metrics_vec
                    else:
                        # recursive case: [bb_idx, subroutine_cost_vec]
                        callee_bb_idx, sub_cost_vec = bb_stack_elem
                        # bb_metrics = bb_cost_trace_df.iloc[callee_bb_idx]
                        # metrics_vec = [int(bb_metrics[ev]) for ev in event_names]
                        # metrics_vec = [event_arrays[ev][callee_bb_idx] for ev in event_names]
                        metrics_vec = event_arrays[:, callee_bb_idx]
                        # cost_vec = [c + m + s for c, m, s in zip(cost_vec, metrics_vec, sub_cost_vec)]
                        cost_vec = cost_vec + metrics_vec + sub_cost_vec

                # print("pop call_stack + bb_stack")
                bb_stack.pop()
                call_stack.pop()
                if (diff - 1) == j:
                    depth = len(call_stack)
                    new_ = [func, depth, first_pc, last_pc, np.zeros(len(event_names), dtype=np.int64)]
                    function_trace.append(new_)

                caller = bb_stack[-1][-1]
                if isinstance(caller, list):
                    caller_bb_idx, caller_cost_vec = caller
                    # record the inclusive cost of the callee relative to caller
                    inclusive_cost_dict[bb_last_pc[caller_bb_idx]][bb_first_pc[callee_first_bb_idx]].append(cost_vec)
                    # update caller’s accumulated cost vector
                    # bb_stack[-1][-1][1] = [c + v for c, v in zip(caller_cost_vec, cost_vec)]
                    bb_stack[-1][-1][1] = caller_cost_vec + cost_vec
                else:
                    caller_bb_idx = caller
                    inclusive_cost_dict[bb_last_pc[caller_bb_idx]][bb_first_pc[callee_first_bb_idx]].append(cost_vec)
                    # replace scalar with [bb_idx, cost_vec]
                    bb_stack[-1][-1] = [caller_bb_idx, cost_vec]

                # When callee function returns, it jumps to the start of the "next" basic block
                # in caller function. For example:
                # 0x6b0  jal    rd, imm <--- in function A
                # ...
                # 0x5ae0 jalr   rd, rs, imm <--- in function B
                # 0x6b4  ... <--- return to function A
                if j == diff - 1:
                    # print("append bb stack")
                    bb_stack[-1].append(bb_idx)

        cost_vec_old = function_trace[-1][-1]
        # metrics_vec = [event_arrays[ev][bb_idx] for ev in event_names]
        metrics_vec = event_arrays[:, bb_idx]
        # cost_vec_new = [c + m for c, m in zip(cost_vec_old, metrics_vec)]
        cost_vec_new = cost_vec_old + metrics_vec
        function_trace[-1][-1] = cost_vec_new
        prev_bb_idx = bb_idx
    # print("ts", ts)
    # import numpy as np
    # tds = np.array(ts)[1:] - np.array(ts)[:-1]
    # print("tds", tds)
    # input()

    # print("inclusive_cost_dict", inclusive_cost_dict)
    # inclusive_cost_dict_sum = {pc: sum([sum(x) for x in costs.values()]) for pc, costs in inclusive_cost_dict.items()}
    # print("inclusive_cost_dict_sum", inclusive_cost_dict_sum)
    # inclusive_cost_dict_total = sum(inclusive_cost_dict_sum.values())
    # print("inclusive_cost_dict_total", inclusive_cost_dict_total)
    # inclusive_cost_dict_max = max(inclusive_cost_dict_sum.values())
    # print("inclusive_cost_dict_sum_max", inclusive_cost_dict_max)
    # print("total_cost", total_cost)
    # TODO: check that all costs are contained
    # print("function_trace", "\n".join(map(str, function_trace[:100])))
    # TODO: export call & func trace (&USE!)
    # function_trace_df = pd.DataFrame(function_trace, columns=["func", "depth", "start_pc", "end_pc", "cost"])
    # print("function_trace_df", function_trace_df)
    # call_trace_df = build_call_trace(function_trace_df, event_names)
    # print("call_trace_df", call_trace_df)
    # input("@")
    # print("inclusive_cost_dict", inclusive_cost_dict)
    # input("%")

    return inclusive_cost_dict


def callgrind_format_converter(
    bb_trace_df: pd.DataFrame,
    unique_bbs_df: pd.DataFrame,
    trace_pcs: Set[int],
    mapping: Dict[str, Tuple[int, int]],
    func2bbs: Dict[str, List[int]],
    srcFile_func_dict: Dict[str, List[str]],
    srcFile_linkage_name_dict: Dict[str, List[str]],
    srcFile_unmangled_linkage_name_dict: Dict[str, List[str]],
    func_set: Set[str],
    file2pc_loc: Optional[Dict[str, List[Tuple[int, int]]]] = None,
    dump_pc: bool = True,
    dump_pos: bool = False,
    elf_file_path: str = "/path/to/elf",
    unmangle_names: bool = False,
    event_names: List[str] = ["Ir", "Cycles"],
    # event_names: List[str] = ["Ir"],
    pcs_hist_df=None,
):
    """
    Generate a callgrind-compatible profile.
    Supports pc-only, line-only, or combined instr+line mode.
    """
    bb_cost_trace_df = pd.DataFrame()
    bb_cost_trace_df["Ir"] = bb_trace_df.merge(unique_bbs_df, how="left", left_on="bb_idx", right_index=True)[
        "num_instrs"
    ]
    bb_cost_trace_df["Cycles"] = bb_cost_trace_df["Ir"] * 2
    # print("bb_cost_trace_df", bb_cost_trace_df)
    # event_arrays = {ev: bb_cost_trace_df[ev].to_numpy() for ev in event_names}
    event_arrays = np.vstack([bb_cost_trace_df[ev].to_numpy() for ev in event_names])

    inclusive_cost_dict = callgrind_format_get_inclusive_cost(
        bb_trace_df, unique_bbs_df, event_names=event_names, event_arrays=event_arrays
    )

    # --- helpers -------------------------------------------------------------
    def find_srcFile(target_func: str) -> str:
        for d in (srcFile_func_dict, srcFile_linkage_name_dict):
            for srcFile, funcs in d.items():
                if target_func in funcs:
                    return srcFile
        return "???"

    def aggregate_pos_cost_of_func(
        sorted_bb_lists: list,
        trace_pcs,
        unique_bbs_df,
        bb_trace_df,
        event_arrays,
        event_names,
        pcs_hist_df: pd.DataFrame,
    ):
        """
        Aggregate per-PC cost vectors for a function.
        Returns: dict pc -> np.ndarray (len(event_names))
        """
        # print("sorted_bb_lists", sorted_bb_lists)
        n_events = len(event_names)
        position_cost_dict = defaultdict(lambda: np.zeros(n_events, dtype=np.int64))
        # bb_trace_arr = bb_trace_df["bb_idx"].to_numpy()

        for bb_idx in sorted_bb_lists:
            bb = unique_bbs_df.iloc[bb_idx]
            pc_start, pc_end = bb.first_pc, bb.last_pc

            # Find all executions of this BB in the trace
            # exec_rows = np.where(bb_trace_arr == bb_idx)[0]

            # for row_idx in exec_rows:
            matching = pcs_hist_df[pcs_hist_df["pc"].isin(range(pc_start, pc_end + 2, 2))]
            # print("matching", matching)
            for row in matching.itertuples():
                pc = row.pc
                count = row.count
                metrics_vec = np.array([count, count * 2])
                position_cost_dict[pc] += metrics_vec
            # if True:
            #     # metrics_vec = np.array([event_arrays[ev][row_idx] for ev in event_names], dtype=np.int64)
            #     # metrics_vec = event_arrays[:, row_idx]

            #     for pc in range(pc_start, pc_end + 2, 2):
            #         if pc not in trace_pcs:
            #             continue
            #         # metrics_vec = np.array([event_arrays[ev][row_idx] for ev in event_names], dtype=np.int64)
            #         # position_cost_dict[pc] += metrics_vec
            #         position_cost_dict[pc] += metrics_vec
        # print("position_cost_dict", position_cost_dict)
        # input("k")
        return position_cost_dict

    def get_line_for_pc(pc: int, source_file: str) -> str:
        if source_file == "???":
            return "0"
        if file2pc_loc is None:
            return "0"
        mapping_ = file2pc_loc.get(source_file)
        if not mapping_:
            return "0"
        mapping_first = [x[0] for x in mapping_]
        i = bisect.bisect_right(mapping_first, pc)
        if i:
            start_pc, line = mapping_[i - 1]
            if pc >= start_pc:
                return str(line)
        return "0"

    # --- positions mode ------------------------------------------------------
    if dump_pc and dump_pos:
        positions = "instr line"
    elif dump_pc:
        positions = "instr"
    else:
        positions = "line"

    events_str = " ".join(event_names)
    prologue = f"""\
# callgrind format
version: 1
creator: callgrind-3.15.0
pid: 2577934
cmd:  ./dhrystone
part: 1

desc: I1 cache:
desc: D1 cache:
desc: LL cache:

desc: Timerange:
desc: Trigger: Program termination

positions: {positions}
events: {events_str}
summary:

"""

    # --- main output ---------------------------------------------------------
    callgrind_output = ""
    srcFile_to_func = defaultdict(list)
    for func in func_set:
        srcFile_to_func[find_srcFile(func)].append(func)

    for srcFile, funcs in srcFile_to_func.items():
        callgrind_output += f"ob={posixpath.abspath(elf_file_path)}\n"
        callgrind_output += f"fl={srcFile}\n"

        for func in funcs:
            # print("func", func)
            bb_list = func2bbs[func]
            bb_list.sort(key=lambda bb_idx: unique_bbs_df.iloc[bb_idx].first_pc)
            if unmangle_names:
                func = unmangle_helper(func)
            callgrind_output += f"fn={func}\n"

            # position_cost_dict = aggregate_pos_cost_of_func(bb_list)
            position_cost_dict = aggregate_pos_cost_of_func(
                bb_list,
                trace_pcs,
                unique_bbs_df,
                bb_trace_df,
                event_arrays,
                event_names,
                pcs_hist_df,
            )
            branch_pc_list = unique_bbs_df.iloc[bb_list].last_pc.values

            for pc in sorted(position_cost_dict.keys()):
                # cost = position_cost_dict[pc]
                costs = position_cost_dict[pc]
                # print("costs", costs)
                costs_str = " ".join(map(str, costs))
                if dump_pc and dump_pos:
                    line = get_line_for_pc(pc, srcFile)
                    callgrind_output += f"{hex(pc)} {line} {costs_str}\n"
                elif dump_pc:
                    callgrind_output += f"{hex(pc)} {costs_str}\n"
                else:  # dump_pos only
                    line = get_line_for_pc(pc, srcFile)
                    callgrind_output += f"{line} {costs_str}\n"

                if pc in branch_pc_list and pc in inclusive_cost_dict:
                    for callee_pc, inclusive_cost in inclusive_cost_dict[pc].items():
                        # print("inclusive_cost", inclusive_cost)
                        # input("123")
                        callgrind_output += f"cob={posixpath.abspath(elf_file_path)}\n"
                        callee_func = find_func_name(mapping, callee_pc)
                        callgrind_output += f"cfi={find_srcFile(callee_func)}\n"
                        if unmangle_names:
                            callee_func = unmangle_helper(callee_func)
                        callgrind_output += f"cfn={callee_func}\n"
                        callgrind_output += f"calls={len(inclusive_cost)} {hex(callee_pc)}\n"
                        total_vec = np.sum(np.stack(inclusive_cost), axis=0)  # shape: (n_events,)
                        sum_str = " ".join(map(str, total_vec))
                        if dump_pc and dump_pos:
                            line = get_line_for_pc(pc, srcFile)
                            # sum all vectors in the list
                            callgrind_output += f"{hex(pc)} {line} {sum_str}\n"
                        elif dump_pc:
                            callgrind_output += f"{hex(pc)} {sum_str}\n"
                        else:
                            line = get_line_for_pc(pc, srcFile)
                            callgrind_output += f"{line} {sum_str}\n"

            callgrind_output += "\n"

    return prologue + callgrind_output


def generate_callgrind_output(
    sess: Session,
    output: Optional[Union[str, Path]] = None,
    force: bool = False,
    dump_pc: bool = False,
    dump_pos: bool = False,
    unmangle_names: bool = False,
):
    artifacts = sess.artifacts

    bb_trace_artifacts = filter_artifacts(artifacts, lambda x: x.name == "bb_trace")
    assert len(bb_trace_artifacts) == 1
    bb_trace_artifact = bb_trace_artifacts[0]
    bb_trace_df = bb_trace_artifact.df

    unique_bbs_artifacts = filter_artifacts(artifacts, lambda x: x.name == "unique_bbs")
    assert len(unique_bbs_artifacts) == 1
    unique_bbs_artifact = unique_bbs_artifacts[0]
    unique_bbs_df = unique_bbs_artifact.df

    func2bbs_artifacts = filter_artifacts(artifacts, lambda x: x.name == "func2bbs")
    assert len(func2bbs_artifacts) == 1
    func2bbs_artifact = func2bbs_artifacts[0]
    func2bbs_df = func2bbs_artifact.df
    func2bbs = dict(zip(func2bbs_df["func"], func2bbs_df["bb_idxs"]))

    trace_pcs_artifacts = filter_artifacts(artifacts, lambda x: x.name == "trace_pcs")
    assert len(trace_pcs_artifacts) == 1
    trace_pcs_artifact = trace_pcs_artifacts[0]
    trace_pcs_df = trace_pcs_artifact.df
    trace_pcs = set(trace_pcs_df.values)  # series?

    elf_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.ELF)
    assert len(elf_artifacts) == 1
    elf_artifact = elf_artifacts[0]

    func2pc_artifacts = filter_artifacts(artifacts, lambda x: x.name == "func2pc")
    assert len(func2pc_artifacts) == 1
    func2pc_artifact = func2pc_artifacts[0]
    func2pc_df = func2pc_artifact.df

    pcs_hist_artifacts = filter_artifacts(artifacts, lambda x: x.name == "pcs_hist")
    assert len(pcs_hist_artifacts) == 1
    pcs_hist_artifact = pcs_hist_artifacts[0]
    pcs_hist_df = pcs_hist_artifact.df
    # print("pcs_hist_df", pcs_hist_df)

    file2funcs_artifacts = filter_artifacts(artifacts, lambda x: x.name == "file2funcs")
    assert len(file2funcs_artifacts) == 1
    file2funcs_artifact = file2funcs_artifacts[0]
    file2funcs_df = file2funcs_artifact.df

    pc2locs_artifacts = filter_artifacts(artifacts, lambda x: x.name == "pc2locs")
    assert len(pc2locs_artifacts) == 1
    pc2locs_artifact = pc2locs_artifacts[0]
    pc2locs_df = pc2locs_artifact.df

    def helper(pc2locs_df):
        df_ = pc2locs_df.explode("locs")
        df_.reset_index(inplace=True)
        df_[["file", "line"]] = df_["locs"].str.split(":", n=1, expand=True)
        ret = {}
        for file, file_df in df_.groupby("file"):
            temp = list(map(tuple, file_df[["pc", "line"]].values))
            temp = sorted(temp, key=lambda x: x[0])
            ret[file] = temp
        return ret

    file2pc_loc = helper(pc2locs_df)

    mapping = func2pc_df.groupby("func")["pc_range"].apply(list).to_dict()
    # bbs, trace_pcs, func2bbs, bb_freq = collect_bbs(trace_artifact.df, mapping)
    # trace_pcs, func2bbs, unique_bbs_df, bb_trace_df = collect_bbs(trace_artifact.df, mapping)

    # bb_trace_attrs = {
    #     "trace": trace_artifact.name,
    #     "kind": "bb_trace",
    #     "by": __name__,
    # }

    # unique_bbs_attrs = {
    #     "trace": trace_artifact.name,
    #     "kind": "mapping",
    #     "by": __name__,
    # }

    # bb_trace_artifact = TableArtifact("bb_trace", bb_trace_df, attrs=bb_trace_attrs)
    # unique_bbs_artifact = TableArtifact("unique_bbs", unique_bbs_df, attrs=unique_bbs_attrs)

    # sess.add_artifact(bb_trace_artifact, override=force)
    # sess.add_artifact(unique_bbs_artifact, override=force)

    file2funcs = dict(list((file2funcs_df[["file", "func_names"]].to_records(index=False))))
    file2linkage_names = dict(list((file2funcs_df[["file", "linkage_names"]].to_records(index=False))))
    file2unmangled_linkage_names = dict(
        list((file2funcs_df[["file", "unmangled_linkage_names"]].to_records(index=False)))
    )
    func_set = set(func2bbs.keys())
    elf_file_path = elf_artifact.path
    # print("M", time.time())

    content = callgrind_format_converter(
        # bbs=bbs,
        bb_trace_df=bb_trace_df,
        unique_bbs_df=unique_bbs_df,
        trace_pcs=trace_pcs,
        mapping=mapping,
        func2bbs=func2bbs,
        # bb_freq=bb_freq,
        srcFile_func_dict=file2funcs,
        srcFile_linkage_name_dict=file2linkage_names,
        srcFile_unmangled_linkage_name_dict=file2unmangled_linkage_names,
        func_set=func_set,
        file2pc_loc=file2pc_loc,
        dump_pc=dump_pc,
        dump_pos=dump_pos,
        elf_file_path=elf_file_path,
        unmangle_names=unmangle_names,
        pcs_hist_df=pcs_hist_df,
    )
    # print("Z", time.time())

    if output is None:
        profile_dir = sess.directory / "profile"
        profile_dir.mkdir(exist_ok=True)
        out_name = "callgrind"
        if dump_pc:
            out_name += "_pc"
        if dump_pos:
            out_name += "_pos"
        out_name += ".out"
        output = profile_dir / out_name
    with open(output, "w") as f:
        f.write(content)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    generate_callgrind_output(
        sess,
        output=args.output,
        force=args.force,
        dump_pc=args.dump_pc,
        dump_pos=args.dump_pos,
        unmangle_names=args.unmangle,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--output", default=None)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--dump-pc", action="store_true")
    parser.add_argument("--dump-pos", action="store_true")
    parser.add_argument("--unmangle", action="store_true")
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import TableArtifact, filter_artifacts
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


def get_unsupported_opcodes(
    allow_mem: bool = False,
    allow_loads: bool = False,
    allow_stores: bool = False,
    allow_branches: bool = False,
    allow_compressed: bool = True,
    allow_custom: bool = True,
    allow_fp: bool = False,
    allow_system: bool = False,
    allow_rvv: bool = False,
):
    unsupported_opcodes = set()

    allow_lui = False
    # allow_lui = True
    if not allow_lui:
        unsupported_opcodes.add("LUI")

    if not allow_compressed:
        unsupported_opcodes.add("LOAD (Compressed)")
        unsupported_opcodes.add("STORE (Compressed)")
        unsupported_opcodes.add("OP-IMM (Compressed)")
        unsupported_opcodes.add("BRANCH (Compressed)")
        unsupported_opcodes.add("MISC-ALU (Compressed)")
        # unsupported_opcodes.add("JAL (Compressed)")
        unsupported_opcodes.add("JALR (Compressed)")
    if not allow_loads or not allow_mem:
        unsupported_opcodes.add("LOAD")
        unsupported_opcodes.add("LOAD-FP")
        unsupported_opcodes.add("LOAD (Compressed)")
    if not allow_stores or not allow_mem:
        unsupported_opcodes.add("STORE")
        unsupported_opcodes.add("STORE-FP")
        unsupported_opcodes.add("STORE (Compressed)")
    if not allow_mem:
        unsupported_opcodes.add("MISC-MEM")
    if not allow_branches:
        unsupported_opcodes.add("BRANCH")
        unsupported_opcodes.add("JAL")
        unsupported_opcodes.add("JALR")
        unsupported_opcodes.add("BRANCH (Compressed)")
        # unsupported_opcodes.add("JAL (Compressed)")
        unsupported_opcodes.add("JALR (Compressed)")
    if not allow_custom:
        unsupported_opcodes.add("custom-0")
        unsupported_opcodes.add("custom-1")
        unsupported_opcodes.add("custom-2/rv128")
        unsupported_opcodes.add("custom-3/rv128")
    if not allow_fp:
        unsupported_opcodes.add("OP-FP")
        unsupported_opcodes.add("LOAD-FP")
        unsupported_opcodes.add("STORE-FP")
    if not allow_rvv:
        unsupported_opcodes.add("OP-V")
    if not allow_system:
        unsupported_opcodes.add("SYSTEM")
    return unsupported_opcodes


def get_ise_potential_df(opcodes_hist_df, unsupported_opcodes, min_supported):
    # print("opcodes_hist_df")
    # print(opcodes_hist_df)
    supported_opcodes_hist_df = opcodes_hist_df[~opcodes_hist_df["opcode"].isin(unsupported_opcodes)]
    # print("supported_opcodes_hist_df")
    # print(supported_opcodes_hist_df)
    total_rel_count = opcodes_hist_df["rel_count"].sum()
    supported_rel_count = supported_opcodes_hist_df["rel_count"].sum() / total_rel_count
    # print("supported_rel_count", supported_rel_count)
    unsupported_rel_count = 1 - supported_rel_count
    has_potential = supported_rel_count >= min_supported
    ise_potential_data = {
        "supported_rel_count": supported_rel_count,
        "unsupported_rel_count": unsupported_rel_count,
        "has_potential": has_potential,
    }
    ise_potential_df = pd.DataFrame([ise_potential_data])
    return ise_potential_df


def check_ise_potential(
    sess: Session,
    min_supported: float = 0.15,
    allow_mem: bool = False,
    allow_loads: bool = False,
    allow_stores: bool = False,
    allow_branches: bool = False,
    allow_compressed: bool = True,
    allow_custom: bool = True,
    allow_fp: bool = False,
    allow_rvv: bool = False,
    allow_system: bool = False,
    force: bool = False,
):
    logger.info("Checking ISE potential...")
    artifacts = sess.artifacts
    opcodes_hist_artifacts = filter_artifacts(artifacts, lambda x: x.name == "opcodes_hist")
    assert len(opcodes_hist_artifacts) == 1
    opcodes_hist_artifact = opcodes_hist_artifacts[0]

    opcodes_hist_df = opcodes_hist_artifact.df

    unsupported_opcodes = get_unsupported_opcodes(
        allow_mem=allow_mem,
        allow_loads=allow_loads,
        allow_stores=allow_stores,
        allow_branches=allow_branches,
        allow_compressed=allow_compressed,
        allow_custom=allow_custom,
        allow_fp=allow_fp,
        allow_rvv=allow_rvv,
        allow_system=allow_system,
    )

    ise_potential_df = get_ise_potential_df(opcodes_hist_df, unsupported_opcodes, min_supported)

    attrs = {
        "kind": "table",
        "by": __name__,
    }

    artifact = TableArtifact("ise_potential", ise_potential_df, attrs=attrs)
    sess.add_artifact(artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    check_ise_potential(
        sess,
        allow_mem=args.allow_mem,
        allow_loads=args.allow_loads,
        allow_stores=args.allow_stores,
        allow_branches=args.allow_branches,
        allow_compressed=args.allow_compressed,
        allow_custom=args.allow_custom,
        allow_fp=args.allow_fp,
        allow_rvv=args.allow_rvv,
        allow_system=args.allow_system,
        force=args.force,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    # TODO: default to allowed?
    parser.add_argument("--min-supported", type=float, default=0.15)
    parser.add_argument("--allow-mem", action="store_true")
    parser.add_argument("--allow-loads", action="store_true")
    parser.add_argument("--allow-stores", action="store_true")
    parser.add_argument("--allow-branches", action="store_true")
    parser.add_argument("--allow-compressed", action="store_true")
    parser.add_argument("--allow-custom", action="store_true")
    parser.add_argument("--allow-fp", action="store_true")
    parser.add_argument("--allow-rvv", action="store_true")
    parser.add_argument("--allow-system", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



import sys
import argparse
from typing import Dict, Optional

import matplotlib.pyplot as plt
import pandas as pd
import yaml


def enc_helper(enc_size, used_space):
    enc_space = 2**32

    major_bits = 7
    minor_bits = 3
    func7_bits = 7
    func2_bits = 2

    major_count = 4  # custom-0, custom-1, custom-2, custom-3
    minor_count = 2**minor_bits
    func2_count = 2**func2_bits
    func7_count = 2**func7_bits

    major_space = 2 ** (enc_size - major_bits)
    minor_space = 2 ** (enc_size - major_bits - minor_bits)
    func2_space = 2 ** (enc_size - major_bits - minor_bits - func2_bits)
    func7_space = 2 ** (enc_size - major_bits - minor_bits - func7_bits)

    major_space_total = major_count * major_space
    minor_space_total = major_count * minor_count * minor_space
    func2_space_total = major_count * minor_count * func2_count * func2_space
    func7_space_total = major_count * minor_count * func7_count * func7_space

    return (
        enc_space,
        used_space,
        major_space,
        minor_space,
        func2_space,
        func7_space,
        major_space_total,
        minor_space_total,
        func2_space_total,
        func7_space_total,
    )


def enc_helper2(properties, enc_size, major_count: int = 4):
    enc_bits_used = properties["OperandEncBitsSum"]
    enc_bits_left = properties[f"EncodingBitsLeft ({enc_size} bits)"]
    enc_footprint = properties[f"EncodingFootprint ({enc_size} bits)"]
    enc_weight = properties[f"EncodingWeight ({enc_size} bits)"]
    assert enc_footprint < 1.0

    # divide weight by 4 as there are 4 major opcodes (custom-0, custom-1, custom-2, custom-3)
    assert major_count in [1, 2, 3, 4]
    enc_weight_ = enc_weight / major_count
    return enc_bits_used, enc_bits_left, enc_footprint, enc_weight_


def enc_helper3(
    used_space,
    major_space,
    major_space_total,
    minor_space,
    minor_space_total,
    func2_space,
    func2_space_total,
    func7_space,
    func7_space_total,
):
    enc_footprint_major = used_space / major_space
    enc_footprint_major_total = used_space / major_space_total

    enc_footprint_minor = used_space / minor_space
    enc_footprint_minor_total = used_space / minor_space_total

    enc_footprint_func2 = used_space / func2_space
    enc_footprint_func2_total = used_space / func2_space_total

    enc_footprint_func7 = used_space / func7_space
    enc_footprint_func7_total = used_space / func7_space_total

    return (
        enc_footprint_major,
        enc_footprint_major_total,
        enc_footprint_minor,
        enc_footprint_minor_total,
        enc_footprint_func2,
        enc_footprint_func2_total,
        enc_footprint_func7,
        enc_footprint_func7_total,
    )


def collect_weights(index_data: dict, enc_size: int = 32, major_count: int = 4):
    total_weight = 0
    weight_per_instr = {}
    bits_per_instr = {}
    footprint_per_instr = {}

    for i, candidate_data in enumerate(index_data["candidates"]):
        # instr_name = f"CUSTOM{i}"
        properties = candidate_data["properties"]
        instr_name = properties["InstrName"]

        # used_space = int(2**enc_bits_used)
        # (enc_space, used_space, major_space, minor_space, func2_space, func7_space, major_space_total,
        #     minor_space_total, func2_space_total, func7_space_total) = enc_helper(enc_size)
        # enc_bits_used, enc_bits_left, enc_footprint, enc_weight
        used_bits, _, enc_footprint, enc_weight = enc_helper2(properties, enc_size, major_count=major_count)
        weight_per_instr[instr_name] = enc_weight
        bits_per_instr[instr_name] = used_bits
        footprint_per_instr[instr_name] = enc_footprint
        total_weight += enc_weight

    rest_weight = 1 - total_weight
    return total_weight, weight_per_instr, footprint_per_instr, rest_weight, bits_per_instr


def plot_enc_pie(
    ax,
    weight_per_instr: Dict[str, float],
    rest_weight: Optional[float] = None,
    combine: bool = False,
    legend: bool = True,
    title: Optional[str] = None,
):
    if combine:
        custom = sum(weight_per_instr.values())
        temp = {"used": custom}
    else:
        temp = {**weight_per_instr}

    if rest_weight is not None:
        temp["free"] = rest_weight

    pie = ax.pie(
        temp.values(),
        labels=list(temp.keys()) if not legend else None,
        autopct="%1.5f%%",
        # legend=legend,
        labeldistance=1 if not legend else None,
    )
    if legend:
        # Matplotlibs hides legend labels starting with an '_'...
        labels = list(temp.keys())
        ax.legend(labels, loc="center left", bbox_to_anchor=(1.0, 0.5))

    if title is not None:
        ax.set_title(title)

        return pie


def plot_enc_pie_multi(weight_per_instr: Dict[str, float], rest_weight: float):

    fig, axs = plt.subplots(1, 2, figsize=(12, 5))

    _ = plot_enc_pie(axs[0], weight_per_instr, rest_weight=rest_weight, combine=True, title="Footprint (Total)")
    _ = plot_enc_pie(axs[1], weight_per_instr, title="Footprint (Custom Only)")

    return fig


def get_enc_weights_df(total_weight, weight_per_instr, bits_per_instr, footprint_per_instr):
    enc_weights_data = [
        {
            "instr": instr,
            "bits": bits_per_instr.get(instr),
            "footprint": footprint_per_instr.get(instr),
            "weight": weight,
        }
        for instr, weight in weight_per_instr.items()
    ] + [{"instr": "*", "weight": total_weight}]
    enc_weights_df = pd.DataFrame(enc_weights_data)
    return enc_weights_df


def get_enc_score_df(enc_weights_df):
    score_df = enc_weights_df.copy().dropna()
    score_df["enc_score"] = score_df.loc[:, ["footprint", "weight"]].apply(
        lambda x: -1 if x["footprint"] > 1.0 else (1.0 - x["weight"]), axis=1
    )
    return score_df


def get_enc_dfs(index_file, enc_size: int = 32, major_count: int = 1):
    with open(index_file, "r") as f:
        combined_index_data = yaml.safe_load(f)

    total_weight, weight_per_instr, footprint_per_instr, rest_weight, bits_per_instr = collect_weights(
        combined_index_data,
        enc_size=enc_size,
        major_count=major_count,
    )
    total_enc_metrics_data = {"total_weight": total_weight}
    total_enc_metrics_df = pd.DataFrame([total_enc_metrics_data])
    enc_weights_df = get_enc_weights_df(total_weight, weight_per_instr, bits_per_instr, footprint_per_instr)
    return total_enc_metrics_df, enc_weights_df


def analyze_encoding(index, enc_size=32, major_count=1, output=None, score=None):
    total_enc_metrics_df, enc_weights_df = get_enc_dfs(index, enc_size=enc_size, major_count=major_count)

    assert output is not None
    total_enc_metrics_df.to_csv(output, index=False)

    if score is not None:
        enc_score_df = get_enc_score_df(enc_weights_df)
        enc_score_df.to_csv(score, index=False)


def handle(args):
    analyze_encoding(
        args.index,
        output=args.output,
        enc_size=args.enc_size,
        major_count=args.major_count,
        score=args.score,
    )


def get_parser():
    parser = argparse.ArgumentParser(description="TODO.")
    parser.add_argument("index", help="Index yaml file")
    parser.add_argument("-o", "--output", required=True, help="Output CSV file")
    parser.add_argument("--enc-size", default=32, help="Encoding Size")
    parser.add_argument("--major-count", type=int, default=1, help="Number of Major Opcodes")
    parser.add_argument("--score", default=None, help="Encoding Score output CSV file")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

from isaac_toolkit.session import Session

# from isaac_toolkit.logging import get_logger, set_log_level

from isaac_toolkit.analysis.dynamic.trace.normalize_trace import normalize_trace

# logger = get_logger()
import logging

logger = logging.getLogger()


def normalize_artifacts(sess: Session, force: bool = False, progress: bool = False):
    logger.info("Normalizing RVF Demo artifacts...")
    normalize_trace(sess, force=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    # set_log_level(console_level=args.log, file_level=args.log)
    normalize_artifacts(sess, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from typing import Union
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


RISCV_OPCODE_MAPPING = {
    0b0010011: "OP-IMM",
    0b0110111: "LUI",
    0b0010111: "AUIPC",
    0b0110011: "OP",
    0b1101111: "JAL",
    0b1100111: "JALR",
    0b1100011: "BRANCH",
    0b0000011: "LOAD",
    0b0100011: "STORE",
    0b0001111: "MISC-MEM",
    0b1110011: "SYSTEM",
    0b1000011: "MADD",
    0b1000111: "MSUB",
    0b1001011: "MNSUB",
    0b1001111: "MNADD",
    0b0000111: "LOAD-FP",
    0b0100111: "STORE-FP",
    0b0001011: "custom-0",
    0b0101011: "custom-1",
    0b1011011: "custom-2/rv128",
    0b1111011: "custom-3/rv128",
    0b1101011: "reserved",
    0b0101111: "AMO",
    0b1010011: "OP-FP",
    0b1010111: "OP-V",
    0b1110111: "OP-P",
    0b0011011: "OP-IMM-32",
    0b0111011: "OP-32",
}

RISCV_COMPRESSED_OPCODE_MAPPING = {
    0b00000: "OP-IMM",
    0b00001: "OP-IMM",
    0b00010: "OP-IMM",
    0b00100: "LOAD",
    0b00101: "JAL",
    0b00110: "LOAD-FP",
    0b01000: "LOAD",
    0b01001: "OP-IMM",
    0b01010: "LOAD",
    0b01100: "LOAD-FP",
    0b01101: "OP-IMM",
    0b01110: "LOAD-FP",
    0b10000: "reserved",
    0b10001: "MISC-ALU",
    0b10010: "JALR",
    0b10100: "STORE-FP",
    0b10101: "JAL",
    0b10110: "STORE-FP",
    0b11000: "STORE",
    0b11001: "BRANCH",
    0b11010: "STORE",
    0b11100: "STORE-FP",
    0b11101: "BRANCH",
    0b11110: "STORE-FP",
}


def decode_opcode(instr_word: Union[str, int]):
    if isinstance(instr_word, str):
        instr_word = int(instr_word, 0)
    elif isinstance(instr_word, float):
        instr_word = int(instr_word)
    assert isinstance(instr_word, int)
    opcode = instr_word & 0b1111111
    lsbs = opcode & 0b11
    if lsbs == 0b11:
        major = RISCV_OPCODE_MAPPING.get(opcode, "UNKNOWN")
        return major
    else:
        # 16-bit instruction
        msbs = (instr_word & 0b1110000000000000) >> 13
        combined = msbs << 2 | lsbs
        assert combined in RISCV_COMPRESSED_OPCODE_MAPPING.keys()
        return f"{RISCV_COMPRESSED_OPCODE_MAPPING[combined]} (Compressed)"


def collect_opcodes(trace_df):
    # Step 1: Count bytecode frequencies directly
    bytecode_counts = trace_df["bytecode"].value_counts()

    # Step 2: Decode each unique bytecode only once
    decoded_map = {bc: decode_opcode(bc) for bc in bytecode_counts.index}

    # Step 3: Replace bytecodes with opcodes and sum counts
    opcode_counts = {}
    for bc, count in bytecode_counts.items():
        op = decoded_map[bc]
        opcode_counts[op] = opcode_counts.get(op, 0) + count

    # Step 4: Build DataFrame
    opcodes_df = pd.DataFrame([{"opcode": op, "count": cnt} for op, cnt in opcode_counts.items()])
    total_count = opcodes_df["count"].sum()
    opcodes_df["rel_count"] = opcodes_df["count"] / total_count
    return opcodes_df.sort_values("count", ascending=False)


def create_opcode_hist(sess: Session, force: bool = False):
    logger.info("Creating opcode historgram...")
    artifacts = sess.artifacts
    trace_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.INSTR_TRACE)
    assert len(trace_artifacts) == 1
    trace_artifact = trace_artifacts[0]

    opcodes_df = collect_opcodes(trace_artifact.df)

    attrs = {
        "trace": trace_artifact.name,
        "kind": "histogram",
        "by": __name__,
    }

    opcodes_artifact = TableArtifact("opcodes_hist", opcodes_df, attrs=attrs)
    sess.add_artifact(opcodes_artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    create_opcode_hist(sess, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# import time
import sys
import pandas as pd
import argparse
from typing import Optional
from pathlib import Path

from tqdm import tqdm

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import TableArtifact


# TODO: logger

ETISS_CPU_TIME_DEFAULT = 0.000000031250


def load_mem_trace(
    sess: Session,
    input_file: Path,
    force: bool = False,
    cpu_time: Optional[float] = None,
):
    assert input_file.is_file()
    name = "mem_trace"
    dfs = []
    with pd.read_csv(
        input_file,
        sep=";",
        names=["time_ps", "pc", "mode", "addr", "bytes"],
        chunksize=2**22,
    ) as reader:
        for df in tqdm(reader, disable=False):
            # df["pc"] = df["pc"].apply(lambda x: int(x, 0))  # TODO: add 0x prefix
            df["pc"] = df["pc"].apply(lambda x: int(x, 16))
            df["pc"] = pd.to_numeric(df["pc"])
            df["addr"] = df["addr"].apply(lambda x: int(x, 16))
            df["addr"] = pd.to_numeric(df["addr"])
            if cpu_time is not None:
                df["idx"] = round(df["time_ps"] * 1e-12 / cpu_time).astype(int)
            # print("B", time.time())
            # TODO: normalize instr names
            df["mode"] = df["mode"].astype("category")

            def chunker(seq, size):
                return (seq[pos : pos + size] for pos in range(0, len(seq), size))

            # for row_df in chunker(df, 16):
            #     print("df.bytes", row_df["bytes"])
            #     print("df.bytes2", row_df["bytes"].astype(int))
            #     print("df.bytes3", row_df["bytes"].apply(lambda x: int(x)))
            #     print("df.bytes4", pd.to_numeric(row_df["bytes"]))
            # print("df.bytes", df["bytes"])
            # print("df.bytes2", df["bytes"].astype(int))
            df["bytes"] = df["bytes"].apply(lambda x: int(str(x), 16))
            # print("df.bytes4", pd.to_numeric(df["bytes"]))
            # df["bytes"] = pd.to_numeric(df["bytes"]).astype("category")  # TODO: does this work?
            df["bytes"] = df["bytes"].astype("category")  # TODO: does this work?
            dfs.append(df)
    df = pd.concat(dfs, axis=0)

    attrs = {
        "simulator": "etiss",
        "cpu_arch": "unknown",
        "by": "isaac_toolkit.frontend.mem_trace.etiss",
    }
    artifact = TableArtifact(name, df, attrs=attrs)
    sess.add_artifact(artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    input_file = Path(args.file)
    load_mem_trace(sess, input_file, force=args.force, cpu_time=args.cpu_time)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("file", help="dBusAccess.csv file generated by ETISS.")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--cpu-time", type=float, default=ETISS_CPU_TIME_DEFAULT)
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from typing import Optional
from pathlib import Path

from isaac_toolkit.session import Session
from isaac_toolkit.logging import get_logger, set_log_level
from isaac_toolkit.retargeting.iss.etiss import retarget_etiss_iss

logger = get_logger()


def retarget_iss_auto(
    sess: Session,
    label: str = "",
    workdir: Optional[str] = None,
    force: bool = False,
    progress: bool = False,
    docker_override: Optional[bool] = None,
    service_override: Optional[bool] = None,
    verbose: bool = False,  # TODO: add to other stages, too
    cleanup: bool = False,
):
    logger.info("Retargeting ISS with ISAAC instructions...")
    # input(">>>")
    config = sess.config
    flow_config = config.flow
    sim = "etiss"
    if flow_config:
        demo_config = flow_config.demo
        if demo_config:
            sim_config = demo_config.sim
            if sim_config:
                sim_ = sim_config.iss
                if sim_:
                    sim = sim_
    assert sim == "etiss"
    sim_retarget_iss = {
        "etiss": retarget_etiss_iss,
    }
    retarget_iss = sim_retarget_iss.get(sim)
    assert retarget_iss is not None
    if workdir is None:
        workdir = sess.directory / "work"  # expose via cmd
    else:
        workdir = Path(workdir)
        assert workdir.is_dir()
    config = sess.config
    flow_config = config.flow
    assert flow_config is not None
    demo_config = flow_config.demo
    assert demo_config is not None
    docker_config = demo_config.docker
    assert docker_config is not None
    service_config = demo_config.service
    assert service_config is not None
    use_docker = docker_override if docker_override is not None else docker_config.enable
    use_service = service_override if service_override is not None else service_config.enable
    etiss_config = demo_config.etiss
    assert etiss_config is not None
    etiss_core = etiss_config.core_name
    mount_dir = None  # Expose to CLI
    if use_docker:
        assert not use_service
        docker_image = docker_config.etiss_image
        if mount_dir is None:
            mount_dir = Path.cwd()
    else:
        docker_image = None
    if use_service:
        assert not use_docker
        service_host = service_config.hostname
        service_port = service_config.port
        service_tags = service_config.tags
        if service_tags is not None:
            service_tag = service_tags.get(sim)
        else:
            service_tag = None
        service_addr = f"{service_host}:{service_port}"
        assert service_tag is not None
    else:
        service_addr = None
        service_tag = None
    # else:
    #     raise NotImplementedError("Non-docker mode")
    retarget_iss(
        sess,
        workdir=workdir,
        mount_dir=mount_dir,
        docker_image=docker_image,
        service_addr=service_addr,
        service_tag=service_tag,
        etiss_core=etiss_core,
        label=label,
        force=force,
        verbose=verbose,
        cleanup=cleanup,
    )


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    docker_override = None
    service_override = False
    if args.docker:
        assert not (args.local or args.service)
        docker_override = True
    if args.local:
        assert not (args.docker or args.servive)
        docker_override = False
    if args.service:
        assert not (args.local or args.docker)
        docker_override = False
        service_override = True
    # TODO: expsoe cleanup via config?
    retarget_iss_auto(
        sess,
        label=args.label,
        workdir=args.workdir,
        force=args.force,
        docker_override=docker_override,
        verbose=args.verbose,
        cleanup=args.cleanup,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--label", default="")
    parser.add_argument("--workdir", default=None)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--docker", action="store_true")
    parser.add_argument("--local", action="store_true")
    parser.add_argument("--service", action="store_true")
    parser.add_argument("--verbose", action="store_true")
    parser.add_argument("--cleanup", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import logging
import argparse
from pathlib import Path

import pandas as pd
import numpy as np

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts, TraceArtifact


logging.basicConfig(level=logging.DEBUG)  # TODO
logger = logging.getLogger(__name__)


def classify(instr):
    instr = instr.lower().replace(".", "_")
    if instr[0] != "v":
        return None
    instr_base = instr.split("_", 1)[0]
    if instr_base in ["vwmacc"]:
        return "MAC"
    if instr_base in ["vmul"]:
        return "MUL"
    if instr_base in ["vadd", "vmin", "vmax", "vxor", "vsll", "vwadd"]:
        return "ARITH"
    if instr.startswith("vset"):
        return "CFG"
    if instr.startswith("vmvr"):
        return "MV Reg"
    if instr == "vmv_s_x":
        return "MV VPR[0] <- GPR"
    if instr == "vmv_v_i":
        return "MV VPR <- IMM"
    if instr == "vmv_v_x":
        return "MV VPR <- GPR"
    if instr == "vmv_x_s":
        return "MV GPR <- VPR[0]"
    if instr.startswith("vmv"):
        return "MV"
    if instr.startswith("vred"):
        return "REDUCE"
    if instr.startswith("vl"):
        return "LOAD"
    if instr.startswith("vsr"):
        return "STORE VREG"
    if instr.startswith("vse") or instr.startswith("vss"):
        return "STORE"
    if instr.startswith("vslide"):
        return "SLIDE"
    # return f"OTHER ({instr})"
    return "OTHER"


def classify_rvv_instrs(instr_trace_df):
    instr_trace_df["instr_alt"] = instr_trace_df["instr"].apply(lambda x: x.lower().replace(".", "_"))
    instr_trace_df["category"] = instr_trace_df["instr_alt"].apply(classify)
    instr_trace_df.drop(columns=["instr_alt"])
    return instr_trace_df


def decode_vsew(x):
    if pd.isna(x):
        return x
    assert int(x) == x
    x = int(x)
    return 8 * 2**x


LMUL_MAP = {
    0b000: "1",
    0b001: "2",
    0b010: "4",
    0b011: "8",
    0b100: "reserved",
    0b101: "1/8",
    0b110: "1/4",
    0b111: "1/2",
}


def decode_vtype(x):
    # print("decode_vtype", x)
    vtype = x["vtype"]
    # print("vtype", vtype)
    if pd.isna(vtype):
        # print("nan")
        return vtype, vtype, vtype, vtype, vtype
    assert int(vtype) == vtype
    vtype = int(vtype)
    vill = (vtype >> 31) & 0b1
    vma = (vtype >> 7) & 0b1
    vta = (vtype >> 6) & 0b1
    vsew = decode_vsew((vtype >> 3) & 0b111)
    vlmul = LMUL_MAP[vtype & 0b111]
    ret = vill, vma, vta, vsew, vlmul
    # print("ret", ret)
    return ret


def decode_rvv_cols(
    perf_trace_df,
):
    rvv_trace_df = perf_trace_df[["vtype", "vl", "vlenb"]]
    # rvv_trace_df[["vill", "vma", "vta", "vsew", "vlmul"]] =
    rvv_trace_df[["vill", "vma", "vta", "vsew", "vlmul"]] = rvv_trace_df.apply(
        decode_vtype, axis=1, result_type="expand"
    )
    rvv_trace_df["vl"] = rvv_trace_df["vl"].replace(0, np.nan)
    rvv_trace_df["vl_util"] = rvv_trace_df["vl"] / rvv_trace_df["vlenb"]
    # print("temp", temp.head())
    # print("rvv_trace_df")
    # print(rvv_trace_df.head())
    vsew_vlmul_hist_df = rvv_trace_df[["vsew", "vlmul"]].value_counts().to_frame()
    vsew_vlmul_hist_df["count_rel"] = vsew_vlmul_hist_df["count"] / len(rvv_trace_df.dropna())
    vl_hist_df = rvv_trace_df["vl"].value_counts().to_frame()
    vl_hist_df["count_rel"] = vl_hist_df["count"] / len(rvv_trace_df.dropna())
    return rvv_trace_df, vsew_vlmul_hist_df, vl_hist_df


def collect_rvv_metrics(
    rvv_trace_df,
    # verbose: bool = False,
):
    num_total_instrs = len(rvv_trace_df)
    num_rvv_instrs = len(rvv_trace_df["vl"].dropna())
    num_rvv_instrs_rel = num_rvv_instrs / num_total_instrs
    mean_vl = rvv_trace_df["vl"].mean()
    min_vl = rvv_trace_df["vl"].min()
    max_vl = rvv_trace_df["vl"].max()
    vlenb = rvv_trace_df["vlenb"].dropna().unique()
    assert len(vlenb) == 1
    vlenb = vlenb[0]
    metrics = {
        "num_total_instrs": num_total_instrs,
        "num_rvv_instrs": num_rvv_instrs,
        "num_rvv_instrs_rel": num_rvv_instrs_rel,
        "mean_vl": mean_vl,
        "min_vl": min_vl,
        "max_vl": max_vl,
        "vlenb": vlenb,
    }

    rvv_metrics_df = pd.DataFrame([metrics])
    return rvv_metrics_df


def get_rvv_metrics(
    sess: Session,
    force: bool = False,
    # verbose: bool = False,
):
    artifacts = sess.artifacts

    instr_trace_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.INSTR_TRACE)
    assert len(instr_trace_artifacts) == 1
    instr_trace_artifact = instr_trace_artifacts[0]
    assert instr_trace_artifact.attrs.get("simulator") in ["etiss_perf", "etiss"]
    instr_trace_df = instr_trace_artifact.df
    perf_trace_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TRACE and x.attrs.get("kind") == "perf_trace"
    )
    assert len(perf_trace_artifacts) == 1
    perf_trace_artifact = perf_trace_artifacts[0]
    assert perf_trace_artifact.attrs.get("simulator") in ["etiss_perf", "etiss"]
    perf_trace_df = perf_trace_artifact.df

    instr_trace_df = classify_rvv_instrs(instr_trace_df.copy())
    categories_hist_df = instr_trace_df["category"].value_counts().to_frame()
    categories_hist_df["count_rel"] = categories_hist_df["count"] / categories_hist_df["count"].sum()
    print("categories", categories_hist_df)
    rvv_trace_df, vsew_vlmul_hist_df, vl_hist_df = decode_rvv_cols(perf_trace_df)
    print(vsew_vlmul_hist_df)
    print(vl_hist_df)

    rvv_metrics_df = collect_rvv_metrics(
        rvv_trace_df,
    )
    print("rvv_metrics_df", rvv_metrics_df)

    attrs = {
        "perf_trace": perf_trace_artifact.name,
        "kind": "table",
        "by": __name__,
    }
    attrs_hist = {
        "perf_trace": perf_trace_artifact.name,
        "kind": "hist",
        "by": __name__,
    }
    attrs2 = {
        "perf_trace": perf_trace_artifact.name,
        "kind": "metrics",
        "by": __name__,
    }

    rvv_trace_artifact = TraceArtifact("rvv_trace", rvv_trace_df, attrs=attrs)
    rvv_metrics_artifact = TableArtifact("rvv_metrics", rvv_metrics_df, attrs=attrs2)
    vsew_vlmul_hist_artifact = TraceArtifact("vsew_vlmul_hist", vsew_vlmul_hist_df, attrs=attrs_hist)
    vl_hist_artifact = TraceArtifact("vl_hist", vl_hist_df, attrs=attrs_hist)
    sess.add_artifact(rvv_trace_artifact, override=force)
    sess.add_artifact(rvv_metrics_artifact, override=force)
    sess.add_artifact(vsew_vlmul_hist_artifact, override=force)
    sess.add_artifact(vl_hist_artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    get_rvv_metrics(sess, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--verbose", "-v", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



import pandas as pd

trace_df = pd.read_pickle("sess/instr_trace/etiss_instrs.log.pkl")
trace_df.drop(columns=["bytecode", "pc", "size"], inplace=True, errors="ignore")
operands_df = pd.read_pickle("sess/instr_trace/etiss_instrs.log.pkl")
df_operands_full = pd.concat([trace_df, operands_df], axis=1)
df_operands_full.drop(columns=["csr", "shamt"], inplace=True, errors="ignore")

# df_short = df.iloc[:10000000]
# df = df_short

df_operands_full["instr"] = df_operands_full["instr"].astype(str)
df_operands_full["pseudo_instr"] = df_operands_full["instr"]  # start with actual instruction
df_operands_full.loc[
    (df_operands_full["instr"] == "addi") & (df_operands_full["rs1"] == 0) & (df_operands_full["imm"] != 0),
    "pseudo_instr",
] = "li"
df_operands_full.loc[
    (df_operands_full["instr"] == "addi") & (df_operands_full["imm"] == 0) & (df_operands_full["rs1"] != 0),
    "pseudo_instr",
] = "mv"

df_operands_full.drop(columns=["imm"], inplace=True, errors="ignore")

df_operands_full["instr"] = df_operands_full["instr"].astype("category")
df_operands_full["pseudo_instr"] = df_operands_full["pseudo_instr"].astype("category")

last_written_instr = {}
rs1_writers = []
rs2_writers = []

for idx, row in df_operands_full.iterrows():
    rs1 = row["rs1"]
    rs2 = row["rs2"]
    rs1_writers.append(last_written_instr.get(rs1, None) if rs1 is not None else None)
    rs2_writers.append(last_written_instr.get(rs2, None) if rs2 is not None else None)
    rd = row["rd"]
    if rd is not None:
        last_written_instr[rd] = row["pseudo_instr"]  # use pseudo name now

df_operands_full["rs1_src"] = rs1_writers
df_operands_full["rs2_src"] = rs2_writers
df_operands_full.rs1_src.value_counts()

df_operands_full["rs1_src_flt"] = df_operands_full["rs1_src"].apply(lambda x: x if x in ["mv", "li"] else "other")
df_operands_full["rs2_src_flt"] = df_operands_full["rs2_src"].apply(lambda x: x if x in ["mv", "li"] else "other")

for instr, instr_df in df_operands_full.groupby("pseudo_instr"):
    num = len(instr_df)
    print(">>>", instr, f"[count={num}]", "<<<")
    rs1_src_counts = instr_df.rs1_src_flt.value_counts()
    rs2_src_counts = instr_df.rs2_src_flt.value_counts()
    rs1_src_counts_rel = rs1_src_counts / num
    rs2_src_counts_rel = rs2_src_counts / num
    if len(rs1_src_counts) > 1:
        print("rs1:", rs1_src_counts_rel.to_dict())
    else:
        print("rs1: -")
    if len(rs2_src_counts) > 1:
        print("rs2:", rs2_src_counts_rel.to_dict())
    else:
        print("rs2: -")



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path
from typing import Optional

import pandas as pd


from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, filter_artifacts

from .report_utils import (
    save_pdf_report,
    JUPYTER_CSS,
    encode_image_base64,
    format_hex,
    monospace_addresses_html,
    monospace_addresses_md,
)


def generate_top_bb_table_html(df: pd.DataFrame, top_n: int = 10) -> str:
    """
    Generate a styled HTML table of the top-N basic blocks sorted by rel_weight.
    """

    def helper(x):
        if x is None:
            return "?"
        if isinstance(x, set):
            return list(x)[0]
        return x

    df["func_name"] = df["func_name"].apply(helper)

    df_sorted = df.sort_values("rel_weight", ascending=False).head(top_n).copy()

    # Format addresses as hex
    df_sorted["start"] = df_sorted["start"].apply(format_hex)
    df_sorted["end"] = df_sorted["end"].apply(format_hex)

    # Select useful columns
    df_display = df_sorted[["func_name", "start", "end", "freq", "size", "num_instrs", "weight", "rel_weight"]]

    def colormap_helper(s):
        # Normalize to 0-1
        norm = (s - s.min()) / (s.max() - s.min())
        # Compute RGB as a gradient from red to green
        colors = [f"rgb({int(255*(1-v))},{int(255*v)},0)" for v in norm]
        # print("colors", colors, len(colors))
        return colors

    # Apply Pandas Styler for a nicer HTML table
    styler = (
        df_display.style.set_caption(f"Top {top_n} Basic Blocks by Relative Weight")
        # .bar(subset=["rel_weight"], color="#4CAF50")
        # .bar(subset=["rel_weight"], color=colormap_helper)
        .bar(subset=["rel_weight"], cmap="coolwarm")
        .format(monospace_addresses_html, subset=["start", "end"])
        .format({"rel_weight": "{:.2%}", "weight": "{:,}", "freq": "{:,}", "size": "{:,}", "num_instrs": "{:,}"})
    )
    return styler.to_html()


def generate_top_bb_table_md(df: pd.DataFrame, top_n: int = 10) -> str:
    """
    Generate a Markdown table of the top-N basic blocks sorted by rel_weight.
    """

    def helper(x):
        if x is None:
            return "?"
        if isinstance(x, set):
            return list(x)[0]
        return x

    df["func_name"] = df["func_name"].apply(helper)

    df_sorted = df.sort_values("rel_weight", ascending=False).head(top_n).copy()

    # Format addresses as hex
    df_sorted["start"] = df_sorted["start"].apply(format_hex)
    df_sorted["end"] = df_sorted["end"].apply(format_hex)

    # Select useful columns
    df_display = df_sorted[["func_name", "start", "end", "freq", "size", "num_instrs", "weight", "rel_weight"]]

    # Format rel_weight as percentage
    df_display["rel_weight"] = df_display["rel_weight"].map(lambda x: f"{x:.2%}")
    df_display[["start", "end"]] = df_display[["start", "end"]].applymap(monospace_addresses_md)

    return df_display.to_markdown(index=False, tablefmt="github")


def collect_plot_files(sess: Session):
    plots_dir = sess.directory / "plots"
    known = [
        "runtime_per_func.jpg",
        "runtime_per_instr.jpg",
        "runtime_per_opcode.jpg",
        "runtime_per_library.jpg",
        "runtime_per_object.jpg",
        "runtime_per_llvm_bb.jpg",
    ]
    return {name: plots_dir / name for name in known if (plots_dir / name).is_file()}


def load_runtime_dfs(sess: Session):
    """Load runtime-related dataframes from artifacts."""
    artifacts = sess.artifacts

    pc2bb_df = None
    instrs_hist_df = None
    opcodes_hist_df = None

    pc2bb = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "pc2bb")
    if pc2bb:
        pc2bb_df = pc2bb[0].df.copy()

    instrs_hist = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "instrs_hist")
    if instrs_hist:
        instrs_hist_df = instrs_hist[0].df.copy()

    opcodes_hist = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "opcodes_hist")
    if opcodes_hist:
        opcodes_hist_df = opcodes_hist[0].df.copy()

    return pc2bb_df, instrs_hist_df, opcodes_hist_df


def generate_runtime_summary(pc2bb_df: Optional[pd.DataFrame]) -> pd.DataFrame:
    """Aggregate runtime per function."""
    if pc2bb_df is None:
        return pd.DataFrame()

    df = pc2bb_df.copy()

    def helper(x):
        if x is None:
            return "?"
        if isinstance(x, set):
            return list(x)[0]
        return x

    df["func_name"] = df["func_name"].apply(helper)
    df = df[["func_name", "rel_weight"]]
    df = df.groupby("func_name", as_index=False).sum()
    df = df.sort_values("rel_weight", ascending=False).reset_index(drop=True)
    return df


def render_report_text(summary_df, instrs_hist_df, opcodes_hist_df, plots, detailed=False) -> str:
    lines = []
    lines.append("# Runtime Report\n")

    if summary_df is not None and not summary_df.empty:
        lines.append("## Runtime per Function\n")
        for _, row in summary_df.iterrows():
            lines.append(f"- {row['func_name']}: {row['rel_weight']:.2%}")
        if "runtime_per_func.jpg" in plots:
            lines.append(f"![Runtime per Function]({plots['runtime_per_func.jpg']})")

    if detailed:
        if instrs_hist_df is not None and not instrs_hist_df.empty:
            lines.append("\n## Runtime per Instruction\n")
            for _, row in instrs_hist_df.sort_values("rel_count", ascending=False).iterrows():
                lines.append(f"- {row['instr']}: {row['rel_count']:.2%}")
            if "runtime_per_instr.jpg" in plots:
                lines.append(f"![Runtime per Instruction]({plots['runtime_per_instr.jpg']})")

        if opcodes_hist_df is not None and not opcodes_hist_df.empty:
            lines.append("\n## Runtime per Opcode\n")
            for _, row in opcodes_hist_df.sort_values("rel_count", ascending=False).iterrows():
                lines.append(f"- {row['opcode']}: {row['rel_count']:.2%}")
            if "runtime_per_opcode.jpg" in plots:
                lines.append(f"![Runtime per Opcode]({plots['runtime_per_opcode.jpg']})")

    return "\n".join(lines)


def render_report_html(
    summary_df, instrs_hist_df, opcodes_hist_df, plots, detailed=False, portable=False, style=False
) -> str:
    parts = []
    parts.append("<html><head>")
    if style:
        parts.append(JUPYTER_CSS)
    parts.append("</head><body>")
    parts.append("<h1>Runtime Report</h1>")

    def img_tag(path: Path, alt: str):
        if portable:
            b64 = encode_image_base64(path)
            return f'<img src="data:image/jpeg;base64,{b64}" alt="{alt}" style="max-width:600px;">'
        else:
            return f'<img src="{path}" alt="{alt}" style="max-width:600px;">'

    if summary_df is not None and not summary_df.empty:
        parts.append("<h2>Runtime per Function</h2>")
        parts.append(summary_df.to_html(index=False, float_format="%.4f", classes="dataframe"))
        if "runtime_per_func.jpg" in plots:
            parts.append(img_tag(plots["runtime_per_func.jpg"], "Runtime per Function"))

    if detailed:
        if instrs_hist_df is not None and not instrs_hist_df.empty:
            parts.append("<h2>Runtime per Instruction</h2>")
            parts.append(instrs_hist_df.to_html(index=False, float_format="%.4f", classes="dataframe"))
            if "runtime_per_instr.jpg" in plots:
                parts.append(img_tag(plots["runtime_per_instr.jpg"], "Runtime per Instruction"))

        if opcodes_hist_df is not None and not opcodes_hist_df.empty:
            parts.append("<h2>Runtime per Opcode</h2>")
            parts.append(opcodes_hist_df.to_html(index=False, float_format="%.4f", classes="dataframe"))
            if "runtime_per_opcode.jpg" in plots:
                parts.append(img_tag(plots["runtime_per_opcode.jpg"], "Runtime per Opcode"))

    parts.append("</body></html>")
    return "\n".join(parts)


def generate_runtime_report(
    sess, output=None, fmt="md", detailed=False, portable=False, style=False, topk=10, force=False
):
    pc2bb_df, instrs_hist_df, opcodes_hist_df = load_runtime_dfs(sess)
    # TODO: by object & by library
    # TODO: disass_counts
    # TODO: mem_footprint
    summary_df = generate_runtime_summary(pc2bb_df)

    # Determine output dir
    out_dir = Path(output) if output else (sess.directory / "reports")
    out_dir.mkdir(exist_ok=True)
    plots = collect_plot_files(sess)

    if fmt in ("md", "txt"):
        body = render_report_text(summary_df, instrs_hist_df, opcodes_hist_df, plots, detailed=detailed)
        if pc2bb_df is not None:
            new = "\n\n## Runtime per Trace BB\n\n"
            new += generate_top_bb_table_md(pc2bb_df, top_n=topk)
            body += new
        ext = "md" if fmt == "md" else "txt"
        outfile = out_dir / f"runtime_report.{ext}"
        outfile.write_text(body, encoding="utf-8")
    elif fmt in ("html", "pdf"):
        body = render_report_html(
            summary_df, instrs_hist_df, opcodes_hist_df, plots, detailed=detailed, portable=portable, style=style
        )
        if pc2bb_df is not None:
            lines = body.splitlines()
            pre = "\n".join(lines[:-1])
            # print("pre", pre, len(pre))
            post = lines[-1]
            # print("post", post, len(post))
            new = "\n<h2>Runtime per Trace BB</h2>\n\n"
            new += generate_top_bb_table_html(pc2bb_df, top_n=topk)
            new = new.replace("<table id=", "<table border='1' class='dataframe dataframe' id=")
            body = pre + new + post
        ext = "html" if fmt == "html" else "pdf"
        outfile = out_dir / f"runtime_report.{ext}"
        if fmt == "html":
            outfile.write_text(body, encoding="utf-8")
        else:
            save_pdf_report(body, outfile)
    else:
        assert False, f"Unsupported fmt: {fmt}"

    print(f"[isaac_toolkit.report] Runtime report written to {outfile}")  # TODO: logging!


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    generate_runtime_report(
        sess,
        output=args.out,
        fmt=args.fmt,
        detailed=args.detailed,
        portable=args.portable,
        style=args.style,
        topk=args.topk,
        force=args.force,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser(description="Generate runtime report from ISAAC session.")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--out", help="Custom output directory (default: SESSION/reports)")
    parser.add_argument("--fmt", choices=["md", "txt", "html", "pdf"], default="md", help="Report format")
    parser.add_argument("--detailed", action="store_true", help="Include detailed breakdowns")
    parser.add_argument("--portable", action="store_true", help="Embed plots as base64 (only for HTML output)")
    parser.add_argument("--style", action="store_true", help="Use custom CSS for HTML output")
    parser.add_argument("--topk", type=int, default=10, help="Limit number of table rows")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from typing import Optional
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import filter_artifacts
from isaac_toolkit.frontend.elf.riscv import load_elf
from isaac_toolkit.frontend.linker_map import load_linker_map
from isaac_toolkit.frontend.instr_trace.etiss import (
    load_instr_trace as load_etiss_instr_trace,
)
from isaac_toolkit.frontend.instr_trace.spike import (
    load_instr_trace as load_spike_instr_trace,
)
from isaac_toolkit.frontend.disass.objdump import load_disass
from isaac_toolkit.frontend.compile_commands.json import load_compile_commands_json
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


# TODO: move to own frontend and share code
def load_mlonmcu_exported_run(sess: Session, run_dir: Path, force: bool = False, progress: bool = False):
    logger.info("Loading MLonMCU exported run...")
    assert run_dir.is_dir(), f"MLonMCU run dir does not exist: {run_dir}"
    elf_file = run_dir / "generic_mlonmcu"
    assert elf_file.is_file(), f"ELF file not found: {elf_file}"
    load_elf(sess, elf_file, force=force)
    linker_map_file = run_dir / "mlif" / "generic" / "linker.map"  # TODO: move to real artifacts?
    if linker_map_file.is_file():
        load_linker_map(sess, linker_map_file, force=force)
    else:
        logger.warning("Skipping loading non-existing linker_map: %s", linker_map_file)
    # TODO: load report via frontend?
    # TODO: load initializer?
    report_file = run_dir / "report.csv"
    assert report_file.is_file()
    report_df = pd.read_csv(report_file)
    assert len(report_df) == 1
    assert "Target" in report_df.columns
    target = report_df["Target"].iloc[0]
    instr_trace_frontends = {
        "etiss": load_etiss_instr_trace,
        "spike": load_spike_instr_trace,
        "spike_rv32": load_spike_instr_trace,
        "spike_rv64": load_spike_instr_trace,
    }
    load_instr_trace = instr_trace_frontends.get(target)
    assert load_instr_trace is not None, f"Frontend lookup failed for target: {target}"
    instr_trace_file = run_dir / f"{target}_instrs.log"
    if instr_trace_file.is_file():
        operands = True  # TODO: store operands in extra artifact!
        load_instr_trace(sess, instr_trace_file, force=force, progress=progress, operands=operands)
    else:
        logger.warning("Skipping loading non-existing instr_trace: %s", instr_trace_file)
    dump_file = run_dir / "generic_mlonmcu.dump"
    if dump_file.is_file():
        load_disass(sess, dump_file, force=force)
    else:
        logger.warning("Skipping loading non-existing disass: %s", dump_file)
    compile_commands_file = run_dir / "mlif" / "compile_commands.json"
    if compile_commands_file.is_file():
        load_compile_commands_json(sess, compile_commands_file, force=force)
    else:
        logger.warning("Skipping loading non-existing compile_commands: %s", compile_commands_file)


def load_mlonmcu_artifacts(sess: Session, run_dir: Path, force: bool = False, progress: bool = False):
    load_mlonmcu_exported_run(sess, run_dir, force=force, progress=progress)


def run_mlonmcu_from_initializer(
    sess: Session,
    home: Optional[Path] = None,
    workdir: Optional[Path] = None,
    until: Optional[str] = None,
    force: bool = False,
    trace_instrs: bool = False,
    progress: bool = False,
):
    logger.info("Running MLonMCU with Initializer...")
    artifacts = sess.artifacts
    initializer_artifacts = filter_artifacts(artifacts, lambda x: x.attrs.get("kind") == "mlonmcu_session_initializer")
    print("initializer_artifacts", initializer_artifacts)
    assert len(initializer_artifacts) == 1
    initializer_artifact = initializer_artifacts[0]
    print("initializer_artifact", initializer_artifact)
    initializer_file = initializer_artifact.path
    print("initializer_file", initializer_file)
    # TODO: docker mode
    # TODO: cmdline mode
    # TODO: check if mlonmcu is installed
    import mlonmcu
    import mlonmcu.context
    from mlonmcu.session.run import RunStage, RunInitializer

    if until is None:
        until = RunStage.RUN
    else:
        until = RunStage[until]

    initializer = RunInitializer.from_file(initializer_file)
    print("initializer", initializer)
    print("initializer.config", initializer.config)
    if trace_instrs:
        if "log_instrs" not in initializer.feature_names:
            initializer.feature_names.append("log_instrs")
        initializer.config["log_instrs.to_file"] = True
    print("initializer.config2", initializer.config)
    # print("initializer.runs", initializer.runs)
    # if len(initializer.runs) > 1:
    #     raise NotImplementedError("Multiple runs not supported")
    if workdir is None:
        workdir = sess.directory / "work" / "local" / "mlonmcu"
        workdir.mkdir(parents=True, exist_ok=True)
    sess_cfg = {
        "session.executor": "process_pool",
    }
    with mlonmcu.context.MlonMcuContext(home, deps_lock="read") as context:
        with context.create_session(dest=workdir, config=sess_cfg) as session:
            session.add_run(initializer, ignore_idx=True)
            assert session.process_runs(
                until=until, context=context, export=True
            ), "Error while processing MLonMCU runs"
            report = session.get_reports()
            print("report", report)
            print("report.df", report.df)
            run_dir = session.results[0].dir
            print("run_dir", run_dir)

    load_mlonmcu_exported_run(sess, run_dir, force=force, progress=progress)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    run_mlonmcu_from_initializer(
        sess,
        force=args.force,
        home=args.home,
        workdir=args.workdir,
        until=args.until,
        trace_instrs=args.trace_instrs,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--home", type=str, default=None)
    parser.add_argument("--workdir", type=str, default=None)
    parser.add_argument("--until", type=str, default=None)
    parser.add_argument("--trace-instrs", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])
